#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of EventListener keeps track of which space on a Page is available
"""
import typing
from decimal import Decimal
from math import ceil
from pathlib import Path

from borb.pdf.canvas.event.begin_page_event import BeginPageEvent
from borb.pdf.canvas.event.chunk_of_text_render_event import ChunkOfTextRenderEvent
from borb.pdf.canvas.event.event_listener import Event, EventListener
from borb.pdf.canvas.event.image_render_event import ImageRenderEvent
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.pdf import PDF


class FreeSpaceFinder(EventListener):
    """
    This implementation of EventListener keeps track of which space on a Page is available
    """

    class Grid:
        """
        This class represents a rasterized, low-res view of a Page.
        But rather than rendering instructions on this raster, each cell simply keeps track
        of whether or not the cell is available. This enables a quick lookup for availability of a given Rectangle.
        """

        def __init__(
            self, page_width: Decimal, page_height: Decimal, resolution: Decimal
        ):
            self._page_width = page_width
            self._page_height = page_height
            self._resolution = resolution
            self._availability: typing.List[typing.List[bool]] = [
                [True for _ in range(0, ceil(self._page_width / self._resolution))]
                for _ in range(0, ceil(self._page_height / self._resolution))
            ]

        def _mark_as_unavailable(self, rectangle: Rectangle) -> "FreeSpaceFinder.Grid":
            x_grid = int(int(rectangle.x) / self._resolution)
            y_grid = int(int(rectangle.y) / self._resolution)
            w = int(int(rectangle.width) / self._resolution)
            h = int(int(rectangle.height) / self._resolution)
            for i in range(x_grid - 1, x_grid + w + 1):
                for j in range(y_grid - 1, y_grid + h + 1):
                    if i < 0 or i >= len(self._availability):
                        continue
                    if j < 0 or j >= len(self._availability[i]):
                        continue
                    self._availability[i][j] = False
            return self

        def _get_free_space(
            self, desired_rectangle: Rectangle
        ) -> typing.Optional[Rectangle]:
            """
            This function returns a Rectangle (or None) of free space (no text rendering operations, no drawing operations) near the given Rectangle
            """
            w = int(int(desired_rectangle.width) / self._resolution)
            h = int(int(desired_rectangle.height) / self._resolution)
            possible_points: typing.List[typing.Tuple[Decimal, Decimal]] = []
            for i in range(0, len(self._availability) - w):
                for j in range(0, len(self._availability[i]) - h):
                    is_free = True
                    for k in range(0, w):
                        for l in range(0, h):
                            if not self._availability[i + k][j + l]:
                                is_free = False
                                break
                        if not is_free:
                            break
                    if is_free:
                        possible_points.append(
                            (
                                Decimal(i * self._resolution),
                                Decimal(j * self._resolution),
                            )
                        )

            # find point closest to desired location
            if len(possible_points) == 0:
                return None
            min_dist = (desired_rectangle.x - possible_points[0][0]) ** 2 + (
                desired_rectangle.y - possible_points[0][1]
            ) ** 2
            min_dist_point = possible_points[0]
            for p in possible_points:
                d = (desired_rectangle.x - p[0]) ** 2 + (
                    desired_rectangle.y - p[1]
                ) ** 2
                if d < min_dist:
                    min_dist = d
                    min_dist_point = p

            # return
            return Rectangle(
                min_dist_point[0],
                min_dist_point[1],
                desired_rectangle.width,
                desired_rectangle.height,
            )

    def __init__(self):
        self._page_number: int = -1
        self._grid_per_page: typing.Dict[int, FreeSpaceFinder.Grid] = {}

    @staticmethod
    def find_free_space_for_page(
        file: Path, page_number: int, desired_rectangle: Rectangle
    ) -> typing.Optional[Rectangle]:
        """
        This function returns the nearest (euclidean distance)
        empty Rectangle that is at least as wide and tall as the
        desired Rectangle.
        If no such Rectangle exists, this method returns None.
        """
        l: FreeSpaceFinder = FreeSpaceFinder()
        with open(file, "rb") as pdf_file_handle:
            PDF.loads(pdf_file_handle, [l])  # type: ignore [arg-type]
        return l.get_free_space_for_page(page_number, desired_rectangle)

    def get_free_space_for_page(
        self, page_number: int, desired_rectangle: Rectangle
    ) -> typing.Optional[Rectangle]:
        """
        This function returns the nearest (euclidean distance)
        empty Rectangle that is at least as wide and tall as the
        desired Rectangle.
        If no such Rectangle exists, this method returns None.
        """
        if page_number in self._grid_per_page:
            return self._grid_per_page[page_number]._get_free_space(desired_rectangle)
        return None

    def _event_occurred(self, event: Event) -> None:

        # BeginPageEvent
        if isinstance(event, BeginPageEvent):
            self._page_number += 1
            self._grid_per_page[self._page_number] = FreeSpaceFinder.Grid(
                event.get_page().get_page_info().get_width() or Decimal(0),
                event.get_page().get_page_info().get_height() or Decimal(0),
                Decimal(10),
            )

        # ChunkOfTextRenderEvent
        if isinstance(event, ChunkOfTextRenderEvent):
            assert isinstance(event, ChunkOfTextRenderEvent)
            bounding_box_001: typing.Optional[
                Rectangle
            ] = event.get_previous_layout_box()
            if bounding_box_001 is not None:
                self._grid_per_page[self._page_number]._mark_as_unavailable(
                    bounding_box_001
                )

        # ImageRenderEvent
        if isinstance(event, ImageRenderEvent):
            assert isinstance(event, ImageRenderEvent)
            bounding_box_002: typing.Optional[Rectangle] = Rectangle(
                event.get_x(), event.get_y(), event.get_width(), event.get_height()
            )
            if bounding_box_002 is not None:
                self._grid_per_page[self._page_number]._mark_as_unavailable(
                    bounding_box_002
                )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Xref tables are part of the original PDF file specification
    and one of the features which gives the PDF file format its flexibility.
    A PDF consists of lots of objects and this tells you where they are located in the file.
    This is actually very useful, as a PDF Reader just has to read these values
    and then it loads the objects only when they are needed.
    It does not need to parse or load the whole file.
"""
import io
import logging
import typing
from decimal import Decimal
from typing import Optional, Union

from borb.io.filter.stream_decode_util import decode_stream
from borb.io.read.tokenize.high_level_tokenizer import HighLevelTokenizer
from borb.io.read.tokenize.low_level_tokenizer import TokenType
from borb.io.read.types import AnyPDFType, Dictionary, Name, Reference, Stream

logger = logging.getLogger(__name__)


class XREF(Dictionary):
    """
    Xref tables are part of the original PDF file specification
    and one of the features which gives the PDF file format its flexibility.
    A PDF consists of lots of objects and this tells you where they are located in the file.
    This is actually very useful, as a PDF Reader just has to read these values
    and then it loads the objects only when they are needed.
    It does not need to parse or load the whole file.
    """

    def __init__(self):
        super(XREF, self).__init__()
        self._entries: typing.List[Reference] = []
        self._cache: typing.Dict[int, Union[AnyPDFType, None]] = {}

    ##
    ## LOWLEVEL IO
    ##

    def _find_startxref_token(
        self,
        src: Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO],
        tok: HighLevelTokenizer,
    ) -> int:

        # measure file length
        src.seek(0, io.SEEK_END)
        file_length: int = src.tell()

        # go to start of search window
        pos: int = max(0, file_length - 1024)
        tok.seek(pos)

        # look for 'startxref'
        while pos > 0:
            # get bytes in window
            bytes_near_eof: bytes = b"".join([tok._next_byte() for _ in range(0, 1024)])
            idx = bytes_near_eof.find(b"startxref")
            if idx >= 0:
                return pos + idx
            # next iteration
            pos = pos - 1024
            tok.seek(pos)

        # not found
        return -1

    def _seek_to_xref_token(
        self,
        src: Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO],
        tok: HighLevelTokenizer,
    ):

        # find "startxref" text
        start_of_xref_token_byte_offset = self._find_startxref_token(src, tok)
        assert start_of_xref_token_byte_offset != -1, "startxref not found in PDF"

        # set tokenizer to "startxref"
        src.seek(start_of_xref_token_byte_offset)
        token = tok.next_non_comment_token()
        assert token is not None
        # if token.get_text() == "xref":
        #    src.seek(start_of_xref_token_byte_offset)
        #    return

        # if we are at startxref, we are reading the XREF table backwards
        # and we need to go back to the start of XREF
        if token.get_text() == "startxref":
            token = tok.next_non_comment_token()
            assert token is not None
            assert token.get_token_type() == TokenType.NUMBER
            start_of_xref_offset = int(token.get_text())
            src.seek(start_of_xref_offset)

    ##
    ## GETTERS AND SETTERS
    ##

    def add(self, r: Reference) -> "XREF":
        """
        Add a new Reference to this XREF
        """
        self._entries.append(r)
        return self

    def merge(self, other_xref: "XREF") -> "XREF":
        """
        Merge this XREF with another XREF
        """
        for r in other_xref._entries:
            duplicate_entries = []
            if r.object_number is not None:
                duplicate_entries = [
                    x for x in self._entries if x.object_number == r.object_number
                ]
            elif r.parent_stream_object_number is not None:
                duplicate_entries = [
                    x
                    for x in self._entries
                    if x.parent_stream_object_number == r.parent_stream_object_number
                    and x.index_in_parent_stream == r.index_in_parent_stream
                ]
            if len(duplicate_entries) == 0:
                self.add(r)
        return self

    def get_object(
        self,
        indirect_reference: Union[Reference, int],
        src: Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO],
        tok: HighLevelTokenizer,
    ) -> Optional[AnyPDFType]:
        """
        This function looks up an object in this XREF table.
        Objects can be looked up by Reference, or object number.
        """
        # cache
        if (
            isinstance(indirect_reference, Reference)
            and indirect_reference.parent_stream_object_number is None
        ):
            assert indirect_reference.object_number is not None
            cached_obj = self._cache.get(indirect_reference.object_number, None)
            if cached_obj is not None:
                return cached_obj

        # lookup Reference object for int
        obj = None
        if isinstance(indirect_reference, int) or isinstance(
            indirect_reference, Decimal
        ):
            refs = [
                x for x in self._entries if x.object_number == int(indirect_reference)
            ]
            if len(refs) == 0:
                return None
            indirect_reference = refs[0]

        # lookup Reference (in self) for Reference
        elif isinstance(indirect_reference, Reference):
            refs = [
                x
                for x in self._entries
                if x.object_number == indirect_reference.object_number
            ]
            if len(refs) == 0:
                return None
            indirect_reference = refs[0]

        # reference points to an object that is not in use
        assert isinstance(indirect_reference, Reference)
        if not indirect_reference.is_in_use:
            obj = None

        # the indirect reference may have a byte offset
        if indirect_reference.byte_offset is not None:
            byte_offset = int(indirect_reference.byte_offset)
            tell_before = tok.tell()
            tok.seek(byte_offset)
            obj = tok.read_object(xref=self)
            tok.seek(tell_before)

        # entry specifies a parent object
        if (
            indirect_reference.parent_stream_object_number is not None
            and indirect_reference.index_in_parent_stream is not None
        ):

            stream_object = self.get_object(
                indirect_reference.parent_stream_object_number, src, tok
            )
            assert isinstance(stream_object, Stream)
            assert "Length" in stream_object
            assert "First" in stream_object

            # Length may be Reference
            if isinstance(stream_object["Length"], Reference):
                stream_object[Name("Length")] = self.get_object(
                    stream_object["Length"], src=src, tok=tok
                )

            # First may be Reference
            if isinstance(stream_object["First"], Reference):
                stream_object[Name("First")] = self.get_object(
                    stream_object["First"], src=src, tok=tok
                )

            first_byte = int(stream_object.get("First", 0))
            if "DecodedBytes" not in stream_object:
                try:
                    stream_object = decode_stream(stream_object)
                except Exception as ex:
                    logger.debug(
                        "unable to inflate stream for object %d"
                        % indirect_reference.parent_stream_object_number
                    )
                    raise ex
            stream_bytes = stream_object["DecodedBytes"][first_byte:]

            # tokenize parent stream
            index = int(indirect_reference.index_in_parent_stream)
            length = int(stream_object["Length"])
            if index < length:
                tok = HighLevelTokenizer(io.BytesIO(stream_bytes))
                list_of_objs = [tok.read_object() for x in range(0, index + 1)]
                obj = list_of_objs[-1]
            else:
                obj = None

        # update cache
        if indirect_reference.parent_stream_object_number is None:
            assert indirect_reference.object_number is not None
            self._cache[indirect_reference.object_number] = obj

        # return
        return obj

    ##
    ## OVERRIDES
    ##

    def __len__(self):
        return len(self._entries)

    def __str__(self):
        out = "xref\n"
        for s in self.sections:
            out += str(s)
        out += "startxref"
        return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of EventListener acts as a filter for other EventListener implementations.
    It only allows events to pass if they occur in a given Rectangle.
"""
import typing

from borb.pdf.canvas.event.chunk_of_text_render_event import ChunkOfTextRenderEvent
from borb.pdf.canvas.event.event_listener import Event, EventListener
from borb.pdf.canvas.event.image_render_event import ImageRenderEvent
from borb.pdf.canvas.geometry.rectangle import Rectangle


class LocationFilter(EventListener):
    """
    This implementation of EventListener acts as a filter for other EventListener implementations.
    It only allows events to pass if they occur in a given Rectangle.
    """

    def __init__(self, rectangle: Rectangle):
        self._rectangle = rectangle
        self._listeners: typing.List[EventListener] = []

    def add_listener(self, listener: "EventListener") -> "LocationFilter":
        """
        This methods add an EventListener to this (meta)-EventListener
        """
        self._listeners.append(listener)
        return self

    def _event_occurred(self, event: "Event") -> None:
        # filter ChunkOfTextRenderEvent
        if isinstance(event, ChunkOfTextRenderEvent):
            for glyph_event in event.split_on_glyphs():
                bb: typing.Optional[Rectangle] = glyph_event.get_previous_layout_box()
                assert bb is not None
                if self._rectangle.x < bb.x < (
                    self._rectangle.x + self._rectangle.width
                ) and self._rectangle.y < bb.y < (
                    self._rectangle.y + self._rectangle.height
                ):
                    for l in self._listeners:
                        l._event_occurred(glyph_event)
            return

        # filter ImageRenderEvent
        if isinstance(event, ImageRenderEvent):
            if self._rectangle.get_x() < event.get_x() < (
                self._rectangle.get_x() + self._rectangle.get_width()
            ) and self._rectangle.get_y() < event.get_y() < (
                self._rectangle.get_y() + self._rectangle.get_height()
            ):
                for l in self._listeners:
                    l._event_occurred(event)
            return

        # default
        for l in self._listeners:
            l._event_occurred(event)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Sashiko (刺し子, lit. "little stabs") is a type of traditional Japanese embroidery or stitching used for the decorative and/or functional reinforcement of cloth and clothing.

Owing to the relatively cheap nature of white cotton thread and the abundant nature of cheap, indigo-dyed blue cloth in historical Japan,
sashiko has a distinctive appearance of white-on-blue embroidery, though some decorative pieces may also use red thread.

First coming into existence in the Edo period (1603-1867), sashiko embroidery was first applied to clothing out of a practical need,
and would have been used to strengthen the homespun clothes of olden times.
Worn out clothes were pieced together to make new garments by using simple running stitches.

These clothes increased their strength with this durable embroidery.
By the Meiji period (1868-1912), sashiko had been established enough that it had evolved into winter work in northern farming communities,
when it was too cold to work outside.

Sashiko was commonly used to reinforce already-patched clothing around points of wear,
but would also be used to attach patches to clothing, making the fabric ultimately stronger.
It would also be used to layer thin fabrics to create warmth, and, in the case of some garments such as the coats of firemen (hikeshibanten),
to create a thick and absorbent material that would be soaked in water before carrying out duties as a fireman.

Though most sashiko utilises only a plain running stitch technique, sashiko is commonly used to create decorative and repeated embroidered patterns,
and may be used for purely decorative purposes, such as in the creation of quilts and embroidery samplers.

Sashiko utilises mostly geometric patterns, which fall into two main styles; moyōzashi,
in which patterns are created with long lines of running stitches; and hitomezashi,
where the pattern emerges from the alignment of single stitches made on a grid.

Common sashiko motifs are waves, mountains, bamboo, arrow feathers, shippō-tsunagi, pampas grass and interlocking geometric shapes,
amongst others; sashiko embroidery is traditionally applied with the use of specialist needles and thread,
though modern day sashiko may use modern embroidery threads and embroidery needles.
"""
import random
import typing
from decimal import Decimal


class RectangularHitomezashi:
    """
    Sashiko (刺し子, lit. "little stabs") is a type of traditional Japanese embroidery or stitching used for the decorative and/or functional reinforcement of cloth and clothing.

    Owing to the relatively cheap nature of white cotton thread and the abundant nature of cheap, indigo-dyed blue cloth in historical Japan,
    sashiko has a distinctive appearance of white-on-blue embroidery, though some decorative pieces may also use red thread.

    First coming into existence in the Edo period (1603-1867), sashiko embroidery was first applied to clothing out of a practical need,
    and would have been used to strengthen the homespun clothes of olden times.
    Worn out clothes were pieced together to make new garments by using simple running stitches.

    These clothes increased their strength with this durable embroidery.
    By the Meiji period (1868-1912), sashiko had been established enough that it had evolved into winter work in northern farming communities,
    when it was too cold to work outside.

    Sashiko was commonly used to reinforce already-patched clothing around points of wear,
    but would also be used to attach patches to clothing, making the fabric ultimately stronger.
    It would also be used to layer thin fabrics to create warmth, and, in the case of some garments such as the coats of firemen (hikeshibanten),
    to create a thick and absorbent material that would be soaked in water before carrying out duties as a fireman.

    Though most sashiko utilises only a plain running stitch technique, sashiko is commonly used to create decorative and repeated embroidered patterns,
    and may be used for purely decorative purposes, such as in the creation of quilts and embroidery samplers.

    Sashiko utilises mostly geometric patterns, which fall into two main styles; moyōzashi,
    in which patterns are created with long lines of running stitches; and hitomezashi,
    where the pattern emerges from the alignment of single stitches made on a grid.

    Common sashiko motifs are waves, mountains, bamboo, arrow feathers, shippō-tsunagi, pampas grass and interlocking geometric shapes,
    amongst others; sashiko embroidery is traditionally applied with the use of specialist needles and thread,
    though modern day sashiko may use modern embroidery threads and embroidery needles.
    """

    @staticmethod
    def hitomezashi(
        horizontal_seed: typing.List[bool] = [
            random.choice([True, False, False]) for _ in range(0, 32)
        ],
        vertical_seed: typing.List[bool] = [
            random.choice([True, False, False]) for _ in range(0, 32)
        ],
    ):
        """
        This function generates a rectangular Hitomezashi grid pattern
        :param horizontal_seed:     a typing.List[bool] to be used as the horizontal seed
        :param vertical_seed:       a typing.List[bool] to be used as the vertical seed
        :return:                    a rectangular Hitomezashi pattern, represented as a typing.List[typing.Tuple[typing.Tuple[Decimal, Decimal], typing.Tuple[Decimal, Decimal]]]
        """

        # keep track of lines
        lines: typing.List[
            typing.Tuple[typing.Tuple[Decimal, Decimal], typing.Tuple[Decimal, Decimal]]
        ] = []

        # build grid
        w: int = len(horizontal_seed)
        h: int = len(vertical_seed)
        for i in range(0, w):
            for j in range(0, h):
                x: Decimal = Decimal(i * 10)
                y: Decimal = Decimal(j * 10)

                # draw line?
                f0: bool = (i % 2 == 0) if vertical_seed[j] else (i % 2 == 1)
                if f0:
                    lines.append(((x, y), (x + 10, y)))

                # draw line?
                f1: bool = (j % 2 == 0) if horizontal_seed[i] else (j % 2 == 1)
                if f1:
                    lines.append(((x, y), (x, y + 10)))

        # return
        return lines



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module defines all the base types used in processing PDFs
e.g. Boolean, CanvasOperatorName, Decimal, Dictionary, Element, Name, Stream, String, ..
"""
import copy
import types
import typing
import xml.etree.ElementTree as ET
from decimal import Decimal as oDecimal
from math import ceil, floor
from typing import Optional, Union

from PIL.Image import Image  # type: ignore [import]

from borb.io.read.postfix.postfix_eval import PostScriptEval


def add_base_methods(object: typing.Any) -> typing.Any:
    """
    This function / decorator adds methods to a given object.
    These added methods are useful for:
    - handling linkage (parent/child relationships),
    - serialization (JSON)
    - hashing
    - implementing the "listener" design pattern
    - etc
    """

    def _to_json_serializable(to_convert=None):
        """
        Convert this object to a representation that
        can be serialized as JSON
        """
        if isinstance(to_convert, dict):
            return {
                to_json_serializable(k): to_json_serializable(v)
                for k, v in to_convert.items()
            }
        if isinstance(to_convert, list):
            return [to_json_serializable(x) for x in to_convert]
        if isinstance(to_convert, Decimal):
            return float(to_convert)
        if (
            isinstance(to_convert, HexadecimalString)
            or isinstance(to_convert, String)
            or isinstance(to_convert, Name)
            or isinstance(to_convert, CanvasOperatorName)
        ):
            return str(to_convert)
        return None

    def to_json_serializable(self):
        """
        This function converts this Object to something that can be JSON serialized
        """
        return _to_json_serializable(self)

    def image_hash_method(self):
        """
        This function hashes Image objects
        """
        w = self.width
        h = self.height
        pixels = [
            self.getpixel((0, 0)),
            self.getpixel((0, h - 1)),
            self.getpixel((w - 1, 0)),
            self.getpixel((w - 1, h - 1)),
        ]
        hashcode = 1
        for p in pixels:
            if isinstance(p, typing.List) or isinstance(p, typing.Tuple):
                hashcode += 32 * hashcode + sum(p)
            else:
                hashcode += 32 * hashcode + p
        return hashcode

    def deepcopy_mod(self, memodict={}):
        """
        This function overrides the __deepcopy__ method
        this was needed
        """
        prev_function_ptr = self.__deepcopy__
        self.__deepcopy__ = None
        # copy
        out = copy.deepcopy(self, memodict)
        # restore
        self.__deepcopy__ = prev_function_ptr
        # add base methods
        add_base_methods(out)
        # return
        return out

    # get parent
    def get_parent(self):
        """
        This function returns the parent Object of the current Object
        """
        if "_parent" not in vars(self):
            setattr(self, "_parent", None)
        return self._parent

    # set parent
    def set_parent(self, parent):
        """
        This function sets the parent Object of the current Object
        """
        if "_parent" not in vars(self):
            setattr(self, "_parent", None)
        self._parent = parent
        return self

    # get_root
    def get_root(self):
        """
        This function returns the root parent Object of the current Object
        """
        e = self
        while e.get_parent() is not None:
            e = e.get_parent()
        return e

    # set_reference
    def set_reference(self, reference: "Reference"):
        """
        This function sets the Reference for this Object, returning self
        """
        if "_reference" not in vars(self):
            setattr(self, "_reference", None)
        assert (
            self._reference is None
            or reference is None
            or self._reference.object_number == reference.object_number
            or (
                self._reference.parent_stream_object_number
                == reference.parent_stream_object_number
                and self._reference.index_in_parent_stream
                == reference.index_in_parent_stream
            )
        )
        self._reference = reference
        return self

    # get_reference
    def get_reference(self) -> typing.Optional["Reference"]:
        """
        This function returns the Reference for this Object or None if no Reference was set
        """
        if "_reference" not in vars(self):
            setattr(self, "_reference", None)
        return self._reference

    # is_uniaue
    def set_is_unique(self, a_flag: bool):
        """
        This function sets whether or not this Object is unique.
        When an object is unique, it is not checked against the cache.
        """
        if "_is_unique" not in vars(self):
            setattr(self, "_is_unique", None)
        self._is_unique = a_flag
        return self

    # is_uniaue
    def get_is_unique(self) -> bool:
        """
        This function returns whether or not this Object is unique.
        When an object is unique, it is not checked against the cache.
        """
        if "_is_unique" not in vars(self):
            setattr(self, "_is_unique", False)
        return self._is_unique

    # is_inline
    def set_is_inline(self, a_flag: bool):
        """
        This function sets whether or not this Object is written inline.
        When an object is inline, it is always embedded immediately in the PDF byte stream.
        """
        if "_is_inline" not in vars(self):
            setattr(self, "_is_inline", None)
        self._is_inline = a_flag
        return self

    # is_inline
    def get_is_inline(self) -> bool:
        """
        This function returns whether or not this Object can be referenced.
        When an object can not be referenced, it is always embedded immediately in the PDF byte stream.
        """
        if "_is_inline" not in vars(self):
            setattr(self, "_is_inline", False)
        return self._is_inline

    object.set_parent = types.MethodType(set_parent, object)
    object.get_parent = types.MethodType(get_parent, object)
    object.get_root = types.MethodType(get_root, object)
    object.set_reference = types.MethodType(set_reference, object)
    object.get_reference = types.MethodType(get_reference, object)
    object.set_is_inline = types.MethodType(set_is_inline, object)
    object.is_inline = types.MethodType(get_is_inline, object)
    object.set_is_unique = types.MethodType(set_is_unique, object)
    object.is_unique = types.MethodType(get_is_unique, object)
    object.to_json_serializable = types.MethodType(to_json_serializable, object)
    if isinstance(object, Image):
        object.__deepcopy__ = types.MethodType(deepcopy_mod, object)
        object.__hash__ = types.MethodType(image_hash_method, object)


class Boolean:
    """
    Boolean objects represent the logical values of true and false. They appear in PDF files using the keywords
    true and false.
    """

    def __init__(self, value: bool):
        super(Boolean, self).__init__()
        self._value = value

    def __bool__(self):
        return self._value

    def __eq__(self, other):
        if isinstance(other, bool):
            return other == self._value
        if isinstance(other, Boolean):
            return other._value == self._value
        return False

    def __str__(self):
        if self._value:
            return "True"
        else:
            return "False"


class CanvasOperatorName:
    """
    This class represents a canvas operator name in PDF syntax
    """

    # fmt: off
    VALID_NAMES = [
        "b", "B", "b*", "B*", "BDC", "BI", "BMC", "BT", "BX",
        "c", "cm", "cs", "CS",
        "d", "d0", "d1", "Do", "DP",
        "EI", "EMC", "ET", "EX",
        "f", "F", "f*",
        "g", "G", "gs",
        "h",
        "i", "ID",
        "j", "J",
        "k", "K",
        "l",
        "m", "M", "MP",
        "n",
        "q", "Q",
        "re", "RG", "rg", "ri",
        "s", "S", "sc", "SC", "SCN", "scn", "sh",
        "T*", "Tc", "Td", "TD", "Tf", "Tj", "TJ", "TL", "Tm", "Tr", "Ts", "Tw", "Tz",
        "v",
        "w", "W", "W*",
        "y",
        "'",
        '"',
    ]
    # fmt: on

    def __init__(self, text: str):
        super(CanvasOperatorName, self).__init__()
        self._text = text
        add_base_methods(self)

    def __eq__(self, other):
        if isinstance(other, CanvasOperatorName):
            return other._text == self._text
        if isinstance(other, str):
            return other == self._text
        return False

    def __hash__(self):
        return self._text.__hash__()

    def __str__(self):
        return self._text


class Decimal(oDecimal):  # type: ignore [no-redef]
    """
    PDF provides two types of numeric objects: integer and real. Integer objects represent mathematical integers.
    Real objects represent mathematical real numbers. The range and precision of numbers may be limited by the
    internal representations used in the computer on which the conforming reader is running; Annex C gives these
    limits for typical implementations.
    """

    def __init__(self, obj: typing.Union[str, float, int, oDecimal]):
        super(Decimal, self).__init__()
        add_base_methods(self)


class Dictionary(dict):
    """
    A dictionary object is an associative table containing pairs of objects, known as the dictionary’s entries. The first
    element of each entry is the key and the second element is the value. The key shall be a name (unlike
    dictionary keys in PostScript, which may be objects of any type). The value may be any kind of object, including
    another dictionary. A dictionary entry whose value is null (see 7.3.9, "Null Object") shall be treated the same as
    if the entry does not exist. (This differs from PostScript, where null behaves like any other object as the value
    of a dictionary entry.) The number of entries in a dictionary shall be subject to an implementation limit; see
    Annex C. A dictionary may have zero entries.

    The entries in a dictionary represent an associative table and as such shall be unordered even though an
    arbitrary order may be imposed upon them when written in a file. That ordering shall be ignored.
    """

    def __init__(self):
        super(Dictionary, self).__init__()
        add_base_methods(self)

    def __hash__(self):
        hashcode: int = 1
        for e in self:
            hashcode = 31 * hashcode + (0 if e is None else hash(e))
        return hashcode

    def __setitem__(self, key, value):
        assert isinstance(key, Name)
        super(Dictionary, self).__setitem__(key, value)

    def __deepcopy__(self, memodict={}):
        out = Dictionary()
        for k, v in self.items():
            out[copy.deepcopy(k, memodict)] = copy.deepcopy(v, memodict)
        return out


class Element(ET.Element):
    """
    An XML element.

    This class is the reference implementation of the Element interface.

    An element's length is its number of subelements.  That means if you
    want to check if an element is truly empty, you should check BOTH
    its length AND its text attribute.

    The element tag, attribute names, and attribute values can be either
    bytes or strings.

    *tag* is the element name.  *attrib* is an optional dictionary containing
    element attributes. *extra* are additional element attributes given as
    keyword arguments.
    """

    def __init__(self, tag, **extra):
        super(Element, self).__init__(tag, **extra)
        add_base_methods(self)


class Name:
    """
    Beginning with PDF 1.2 a name object is an atomic symbol uniquely defined by a sequence of any characters
    (8-bit values) except null (character code 0). Uniquely defined means that any two name objects made up of
    the same sequence of characters denote the same object. Atomic means that a name has no internal structure;
    although it is defined by a sequence of characters, those characters are not considered elements of the name.
    """

    def __init__(self, text: str):
        self._text = text
        add_base_methods(self)

    def __eq__(self, other):
        if isinstance(other, Name):
            return other._text == self._text
        if isinstance(other, str):
            return other == self._text
        return False

    def __hash__(self):
        return self._text.__hash__()

    def __str__(self):
        return self._text


class Stream(Dictionary):
    """
    A stream object, like a string object, is a sequence of bytes. Furthermore, a stream may be of unlimited length,
    whereas a string shall be subject to an implementation limit. For this reason, objects with potentially large
    amounts of data, such as images and page descriptions, shall be represented as streams.
    """

    def __init__(self):
        super(Stream, self).__init__()


class Function(Dictionary):
    """
    A function object may be a dictionary or a stream, depending on the type of function. The term function
    dictionary is used generically in this sub-clause to refer to either a dictionary object or the dictionary portion of a
    stream object. A function dictionary specifies the function’s representation, the set of attributes that
    parameterize that representation, and the additional data needed by that representation. Four types of
    functions are available, as indicated by the dictionary’s FunctionType entry.
    """

    def __init__(self):
        super(Function, self).__init__()

    @staticmethod
    def _interpolate(
        x: oDecimal, x_min: oDecimal, x_max: oDecimal, y_min: oDecimal, y_max: oDecimal
    ) -> oDecimal:
        return y_min + (x - x_min) * ((y_max - y_min) / (x_max - x_min))

    def _get_sample_number(self, sample: typing.List[oDecimal]) -> typing.Optional[int]:
        size: typing.List[int] = [int(x) for x in self["Size"]]
        N: int = 1
        for s in size:
            N *= s
        F: typing.List[int] = [int(N / size[0])]
        for i in range(1, len(size)):
            F[i] = int(F[i - 1] / size[i])
        return sum([F[i] * int(sample[i]) for i in range(0, len(sample))])

    def _get_sample(self, sample_number: int) -> typing.List[oDecimal]:
        n: int = int(len(self["Range"]) / 2)
        bps: int = int(self["BitsPerSample"])
        byte_start_index: int = floor((sample_number * bps * n) / 8)
        byte_stop_index: int = min(
            byte_start_index + ceil((n * bps) / 8), len(self["DecodedBytes"])
        )
        bit_offset: int = (sample_number * bps * n) - byte_start_index * 8
        bytes_to_use: bytes = self["DecodedBytes"][byte_start_index:byte_stop_index]
        byte_str: str = "".join([bin(x)[2:].zfill(8) for x in bytes_to_use])[
            bit_offset : (bit_offset + n * bps)
        ]
        ys: typing.List[oDecimal] = [
            Decimal(int(byte_str[i : i + bps], 2)) for i in range(0, len(byte_str), bps)
        ]
        return ys

    def evaluate(self, xs: typing.List[oDecimal]) -> typing.List[oDecimal]:
        """
        This function evaluates this Function in the given arguments, returning a typing.List[Decimal] as output
        """
        # Type 0 functions use a sequence of sample values (contained in a stream) to provide an approximation for
        # functions whose domains and ranges are bounded. The samples are organized as an m-dimensional table in
        # which each entry has n components.
        if "FunctionType" in self and int(self["FunctionType"]) == 0:
            size: typing.List[oDecimal] = self["Size"]
            bps: int = int(self["BitsPerSample"])
            m: int = len(size)
            domain: typing.List[oDecimal] = self["Domain"]
            encode: typing.List[oDecimal] = []
            decode: typing.List[oDecimal] = []
            range2: typing.List[oDecimal] = self["Range"]
            if "Encode" in self:
                encode = self["Encode"]
            else:
                encode = [
                    oDecimal(0) if (i % 2 == 0) else size[int((i - 1) / 2)]
                    for i in range(0, 2 * m)
                ]
            if "Decode" in self:
                decode = self["Decode"]
            else:
                decode = range2

            # When a sampled function is called, each input value xi , for 0 £ i < m, shall be clipped to the domain:
            xs_prime = [
                min(max(xs[i], domain[2 * i]), domain[2 * i + 1]) for i in range(0, m)
            ]

            # That value shall be encoded:
            es = [
                Function._interpolate(
                    xs_prime[i],
                    domain[2 * i],
                    domain[2 * i + 1],
                    encode[2 * i],
                    encode[2 * i + 1],
                )
                for i in range(0, m)
            ]

            # That value shall be clipped to the size of the sample table in that dimension:
            es_prime = [
                min(max(es[i], oDecimal(0)), size[i] - 1) for i in range(0, len(es))
            ]

            # The encoded input values shall be real numbers, not restricted to integers. Interpolation shall be used to
            # determine output values from the nearest surrounding values in the sample table.
            sample_number: typing.Optional[int] = self._get_sample_number(es_prime)
            assert sample_number is not None
            rs = self._get_sample(sample_number)

            # Each output value rj, for 0 £ j < n, shall then be decoded:
            rs_prime: typing.List[oDecimal] = [
                Function._interpolate(
                    rs[j], oDecimal(0), 2**bps - 1, decode[2 * j], decode[2 * j + 1]
                )
                for j in range(0, len(rs))
            ]

            # Finally, each decoded value shall be clipped to the range:
            ys = [
                min(max(rs_prime[j], range2[2 * j]), range2[2 * j + 1])
                for j in range(0, len(rs_prime))
            ]

            # return
            return ys

        # Type 2 functions (PDF 1.3) include a set of parameters that define an exponential interpolation of one input
        # value and n output values:
        if "FunctionType" in self and int(self["FunctionType"]) == 2:
            assert len(xs) == 1
            if xs[0] == oDecimal(0):
                return self["C0"]
            if xs[0] == Decimal(1):
                return self["C1"]
            n: int = len(self["C0"])
            N: oDecimal = self["N"]
            c0: typing.List[oDecimal] = self["C0"]
            c1: typing.List[oDecimal] = self["C1"]
            return [(c0[j] + xs[0] ** N * (c1[j] - c0[j])) for j in range(0, n)]

        if "FunctionType" in self and int(self["FunctionType"]) == 3:
            # TODO : implement stitching function
            pass

        if "FunctionType" in self and int(self["FunctionType"]) == 4:
            return PostScriptEval.evaluate(self["DecodedBytes"].decode("latin1"), xs)

        # this should be impossible
        assert False

    def __deepcopy__(self, memodict={}):
        out: Function = Function()
        for k, v in self.items():
            out[k] = copy.deepcopy(v, memodict)
        return out


class String:
    """
    A literal string shall be written as an arbitrary number of characters enclosed in parentheses. Any characters
    may appear in a string except unbalanced parentheses (LEFT PARENHESIS (28h) and RIGHT
    PARENTHESIS (29h)) and the backslash (REVERSE SOLIDUS (5Ch)), which shall be treated specially as
    described in this sub-clause. Balanced pairs of parentheses within a string require no special treatment.
    """

    def __init__(self, bts: typing.Union[bytes, str]):  # type: ignore [name-defined]
        if isinstance(bts, str):
            self._text: str = bts
        if isinstance(bts, bytes):
            self._text = [(b & 0xFF) for b in bts]
        add_base_methods(self)

    def __eq__(self, other):
        if isinstance(other, String):
            return other._text == self._text
        if isinstance(other, str):
            return other == self._text
        return False

    def __hash__(self):
        return self._text.__hash__()

    def __str__(self):
        return self._text

    def __len__(self):
        return len(self._text)

    def __getitem__(self, item):
        return self._text[item]

    def get_content_bytes(self) -> bytearray:
        """
        This function returns the bytes that represent the *interpreted* content (as it was present in the PDF)
        of this String. For hexadecimal Strings, this content includes resolving the hexadecimal codes to their
        byte counterparts.
        """
        txt = ""
        i = 0
        while i < len(self):
            if self[i] == "\\":
                c = self[i + 1]
                if c == "n":
                    txt += "\n"
                elif c == "\\":
                    txt += "\\"
                elif c == "r":
                    txt += "\r"
                elif c == "t":
                    txt += "\t"
                elif c == "b":
                    txt += "\b"
                elif c == "f":
                    txt += "\f"

                elif c == "(" or c == ")" or c == "\\":
                    txt += c
                    i += 2
                    continue

                elif c == "\r":
                    if i + 2 < len(self) and self[i + 2] == "\n":
                        i += 3
                    else:
                        i += 2
                    continue

                elif c == "\n":
                    i += 2
                    continue

                else:
                    # We have read <SLASH>
                    # Is the next character <OCTAL> ?
                    # YES:  continue reading as <OCTAL> (max 2 more chars)
                    # NO:   do not process next char
                    if c < "0" or c > "7":
                        txt += c
                        i += 2  # processed <SLASH> <any> (pretend it's a useless escape sequence)
                        continue

                    # we have read <SLASH> <OCTAL>
                    # Is the next character <OCTAL> ?
                    # YES:  continue reading <OCTAL> (max 1 more char)
                    # NO:   do not process next char
                    octal = ord(c) - ord("0")
                    c = self[i + 2]

                    if c < "0" or c > "7":
                        txt += chr(octal)
                        i += 2  # processed <SLASH> <OCTAL>
                        continue

                    # we have read <SLASH> <OCTAL> <OCTAL>
                    octal = (octal << 3) + ord(c) - ord("0")
                    c = self[i + 3]
                    if c < "0" or c > "7":
                        txt += chr(octal)
                        i += 3  # processed <SLASH> <OCTAL> <OCTAL>
                        continue

                    # we have read <SLASH> <OCTAL> <OCTAL> <OCTAL>
                    octal = (octal << 3) + ord(c) - ord("0")
                    txt += chr(octal)
                    i += 4
                    continue

                i += 2
                continue
            txt += self[i]
            i += 1
        return bytearray(txt, encoding="latin1")

    def get_value_bytes(self):
        """
        This function returns the bytes that represent the content (as it was present in the PDF)
        of this String
        """
        return [b for b in self.get_content_bytes()]


class HexadecimalString(String):
    """
    Strings may also be written in hexadecimal form, which is useful for including arbitrary binary data in a PDF file.
    A hexadecimal string shall be written as a sequence of hexadecimal digits (0–9 and either A–F or a–f) encoded
    as ASCII characters and enclosed within angle brackets (using LESS-THAN SIGN (3Ch) and GREATER-
    THAN SIGN (3Eh)).
    """

    def __init__(self, text: str, encoding: Optional["Encoding"] = None):  # type: ignore [name-defined]
        if len(text) % 2 == 1:
            text += "0"
        self.encoding = encoding
        super(HexadecimalString, self).__init__(text)

    def get_content_bytes(self) -> bytearray:
        """
        This function returns the bytes that represent the *interpreted* content (as it was present in the PDF)
        of this String. For hexadecimal Strings, this content includes resolving the hexadecimal codes to their
        byte counterparts.
        """
        arr = bytearray()
        for i in range(0, len(self), 2):
            arr.append(int(self[i : i + 2], 16))
        return arr


class List(list):
    """
    An array object is a one-dimensional collection of objects arranged sequentially. Unlike arrays in many other
    computer languages, PDF arrays may be heterogeneous; that is, an array’s elements may be any combination
    of numbers, strings, dictionaries, or any other objects, including other arrays. An array may have zero
    elements.
    """

    def __init__(self):
        super(List, self).__init__()
        add_base_methods(self)

    def __hash__(self):
        hashcode: int = 1
        for e in self:
            hashcode = 31 * hashcode + (0 if e is None else hash(e))
        return hashcode


class Reference:
    """
    Any object in a PDF file may be labelled as an indirect object. This gives the object a unique object identifier by
    which other objects can refer to it (for example, as an element of an array or as the value of a dictionary entry).

    The object identifier shall consist of two parts:

    •   A positive integer object number. Indirect objects may be numbered sequentially within a PDF file, but this
        is not required; object numbers may be assigned in any arbitrary order.

    •   A non-negative integer generation number. In a newly created file, all indirect objects shall have generation
        numbers of 0. Nonzero generation numbers may be introduced when the file is later updated; see sub-
        clauses 7.5.4, "Cross-Reference Table" and 7.5.6, "Incremental Updates."

    Together, the combination of an object number and a generation number shall uniquely identify an indirect
    object.
    """

    object_number: Optional[int]
    generation_number: Optional[int]
    parent_stream_object_number: Optional[int]
    index_in_parent_stream: Optional[int]
    byte_offset: Optional[int]
    is_in_use: bool
    document: "Document"  # type: ignore [name-defined]

    def __init__(
        self,
        object_number: Optional[int] = None,
        generation_number: Optional[int] = None,
        parent_stream_object_number: Optional[int] = None,
        index_in_parent_stream: Optional[int] = None,
        byte_offset: Optional[int] = None,
        is_in_use: bool = True,
        document: Optional["Document"] = None,  # type: ignore [name-defined]
    ):
        self.object_number = object_number
        self.generation_number = generation_number
        self.parent_stream_object_number = parent_stream_object_number
        self.index_in_parent_stream = index_in_parent_stream
        self.byte_offset = byte_offset
        self.is_in_use = is_in_use
        self.document = document
        add_base_methods(self)

    def __hash__(self):
        hashcode: int = 1
        hashcode = hashcode * 31 + (
            self.object_number if self.object_number is not None else 0
        )
        hashcode = hashcode * 31 + (
            self.generation_number if self.generation_number is not None else 0
        )
        hashcode = hashcode * 31 + (
            self.parent_stream_object_number
            if self.parent_stream_object_number is not None
            else 0
        )
        hashcode = hashcode * 31 + (
            self.index_in_parent_stream
            if self.index_in_parent_stream is not None
            else 0
        )
        hashcode = hashcode * 31 + (
            self.byte_offset if self.byte_offset is not None else 0
        )
        return hashcode

    def __eq__(self, other):
        if not isinstance(other, Reference):
            return False
        return (
            self.object_number == other.object_number
            and self.generation_number == other.generation_number
            and self.parent_stream_object_number == other.parent_stream_object_number
            and self.index_in_parent_stream == other.index_in_parent_stream
            and self.byte_offset == other.byte_offset
        )


AnyPDFType = Union[
    Boolean,
    CanvasOperatorName,
    Decimal,
    Dictionary,
    Element,
    HexadecimalString,
    Name,
    List,
    Reference,
    String,
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Move to the next line and show a text string, using a w as the word spacing
and a c as the character spacing (setting the corresponding parameters in
the text state). a w and a c shall be numbers expressed in unscaled text
space units.
"""
import typing

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class SetSpacingMoveToNextLineShowText(CanvasOperator):
    """
    Move to the next line and show a text string, using a w as the word spacing
    and a c as the character spacing (setting the corresponding parameters in
    the text state). a w and a c shall be numbers expressed in unscaled text
    space units. This operator shall have the same effect as this code:
    aw Tw
    ac Tc
    string '
    """

    def __init__(self):
        super().__init__('"', 3)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the " operator
        """
        set_word_spacing_op: typing.Optional[
            CanvasOperator
        ] = canvas_stream_processor.get_operator("Tw")
        assert (
            set_word_spacing_op
        ), 'Operator Tw must be defined for operator " to function'
        set_word_spacing_op.invoke(
            canvas_stream_processor, [operands[0]], event_listeners
        )

        set_character_spacing_op: typing.Optional[
            CanvasOperator
        ] = canvas_stream_processor.get_operator("Tc")
        assert (
            set_character_spacing_op
        ), 'Operator Tc must be defined for operator " to function'
        set_character_spacing_op.invoke(
            canvas_stream_processor, [operands[1]], event_listeners
        )

        move_to_next_line_show_text_op: typing.Optional[
            CanvasOperator
        ] = canvas_stream_processor.get_operator("'")
        assert (
            move_to_next_line_show_text_op
        ), "Operator ' must be defined for operator \" to function"
        move_to_next_line_show_text_op.invoke(
            canvas_stream_processor, [operands[2]], event_listeners
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of EventListener extracts all paragraphs of text from a PDF Document
"""
import typing
from decimal import Decimal

from borb.datastructure.disjoint_set import disjointset
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.layout_element import LayoutElement
from borb.pdf.canvas.layout.text.paragraph import Paragraph
from borb.pdf.page.page import Page
from borb.toolkit.text.simple_line_of_text_extraction import SimpleLineOfTextExtraction


class SimpleParagraphExtraction(SimpleLineOfTextExtraction):
    """
    This implementation of EventListener extracts all paragraphs of text from a PDF Document
    """

    def __init__(
        self,
        minimum_horizontal_overlap_percentage: Decimal = Decimal(0.80),
        maximum_multiplied_leading: Decimal = Decimal(1.40),
    ):
        super(SimpleParagraphExtraction, self).__init__()
        self._minimum_horizontal_overlap_percentage = (
            minimum_horizontal_overlap_percentage
        )
        self._maximum_multiplied_leading = maximum_multiplied_leading
        self._paragraphs_per_page: typing.Dict[int, typing.List[Paragraph]] = {}

    def _end_page(self, page: Page):
        super(SimpleParagraphExtraction, self)._end_page(page)

        # build initial disjointset
        line_of_text_disjoint_set = disjointset()
        for line_of_text in self._lines_of_text_per_page[self._current_page_number]:
            line_of_text_disjoint_set.add(line_of_text)

        # merge all LineOfText objects that represent a line of text
        for l0 in line_of_text_disjoint_set:
            for l1 in line_of_text_disjoint_set:
                if l0 == l1:
                    continue
                if line_of_text_disjoint_set.find(l0) == line_of_text_disjoint_set.find(
                    l1
                ):
                    continue

                if (
                    l0.get_previous_layout_box().get_width() == 0
                    or l1.get_previous_layout_box().get_width() == 0
                ):
                    continue

                # determine overlap
                overlap_percentage = self._overlap(
                    l0.get_previous_layout_box(), l1.get_previous_layout_box()
                ) / min(
                    l0.get_previous_layout_box().width,
                    l1.get_previous_layout_box().width,
                )

                # determine leading
                leading = abs(
                    l0.get_previous_layout_box().y - l1.get_previous_layout_box().y
                ) / min(
                    l0.get_previous_layout_box().height,
                    l1.get_previous_layout_box().height,
                )

                if (
                    overlap_percentage >= self._minimum_horizontal_overlap_percentage
                    and leading <= self._maximum_multiplied_leading
                ):
                    line_of_text_disjoint_set.union(l0, l1)
                    break

        # combine partitions into Paragraph objects
        paragraphs: typing.List[Paragraph] = []
        for line_of_text_partition in line_of_text_disjoint_set.sets():
            lines_of_text = [x for x in line_of_text_partition]

            # determine text
            txt = "".join([x._text + "\n" for x in lines_of_text])[:-1]

            # create / append paragraph
            p: LayoutElement = Paragraph(
                text=txt,
                font=lines_of_text[0]._font,
                font_color=lines_of_text[0]._font_color,
                font_size=lines_of_text[0]._font_size,
            )
            p._previous_layout_box = Rectangle(
                min([l.get_previous_layout_box().x for l in lines_of_text]),
                min([l.get_previous_layout_box().y for l in lines_of_text]),
                max(
                    [
                        l.get_previous_layout_box().x
                        + l.get_previous_layout_box().width
                        for l in lines_of_text
                    ]
                )
                - min([l.get_previous_layout_box().x for l in lines_of_text]),
                max(
                    [
                        l.get_previous_layout_box().y
                        + l.get_previous_layout_box().height
                        for l in lines_of_text
                    ]
                )
                - min([l.get_previous_layout_box().y for l in lines_of_text]),
            )
            assert isinstance(p, Paragraph)
            paragraphs.append(p)

        # add to dict
        self._paragraphs_per_page[self._current_page_number] = paragraphs

    def get_paragraphs_for_page(self, page: int) -> typing.List[Paragraph]:
        """
        This function returns the paragraphs on a given page
        """
        return self._paragraphs_per_page.get(page, [])

    def _overlap(self, r0: Rectangle, r1: Rectangle) -> Decimal:
        # rectangles do not overlap (r0 is on the left side)
        if max(r0.x, r0.x + r0.width) < min(r1.x, r1.x + r1.width):
            return Decimal(0)
        # rectangles do not overlap (r1 is on the left side)
        if max(r1.x, r1.x + r1.width) < min(r0.x, r0.x + r0.width):
            return Decimal(0)
        a = max(min(r0.x, r0.x + r0.width), min(r1.x, r1.x + r1.width))
        b = min(max(r0.x, r0.x + r0.width), max(r1.x, r1.x + r1.width))
        return abs(a - b)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of LayoutElement represents an ordered (that is to say numbered) list.
For this list, roman numerals are used.
"""

from borb.pdf.canvas.color.color import X11Color
from borb.pdf.canvas.layout.layout_element import LayoutElement
from borb.pdf.canvas.layout.list.ordered_list import OrderedList
from borb.pdf.canvas.layout.text.chunk_of_text import ChunkOfText


class RomanNumeralOrderedList(OrderedList):
    """
    This implementation of LayoutElement represents an ordered (that is to say numbered) list.
    For this list, roman numerals are used.
    """

    @staticmethod
    def _int_to_roman(value: int) -> str:
        """Convert an integer to a Roman numeral."""
        assert value > 0, "_int_to_roman can only convert values from 1 to 3999"
        assert value < 4000, "_int_to_roman can only convert values from 1 to 3999"
        ints = (1000, 900, 500, 400, 100, 90, 50, 40, 10, 9, 5, 4, 1)
        nums = ("M", "CM", "D", "CD", "C", "XC", "L", "XL", "X", "IX", "V", "IV", "I")
        result = []
        for i in range(len(ints)):
            count = int(value / ints[i])
            result.append(nums[i] * count)
            value -= ints[i] * count
        return "".join(result)

    def _get_bullet_layout_element(
        self, item_index: int, item: LayoutElement
    ) -> LayoutElement:
        return ChunkOfText(
            text=self._int_to_roman(item_index + 1) + ".",
            font_size=self.get_font_size(),
            font_color=X11Color("Black"),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A sound annotation (PDF 1.2) shall analogous to a text annotation except that instead of a text note, it contains
sound recorded from the computer’s microphone or imported from a file. When the annotation is activated, the
sound shall be played. The annotation shall behave like a text annotation in most ways, with a different icon (by
default, a speaker) to indicate that it represents a sound. Table 185 shows the annotation dictionary entries
specific to this type of annotation. Sound objects are discussed in 13.3, “Sounds.”
"""
from pathlib import Path

from borb.io.read.types import Name, Dictionary, String, Boolean, Decimal as bDecimal
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.annotation.annotation import Annotation


class SoundAnnotation(Annotation):
    """
    A sound annotation (PDF 1.2) shall analogous to a text annotation except that instead of a text note, it contains
    sound recorded from the computer’s microphone or imported from a file. When the annotation is activated, the
    sound shall be played. The annotation shall behave like a text annotation in most ways, with a different icon (by
    default, a speaker) to indicate that it represents a sound. Table 185 shows the annotation dictionary entries
    specific to this type of annotation. Sound objects are discussed in 13.3, “Sounds.”
    """

    @staticmethod
    def _make_canonical_file_path(p: str) -> str:
        try:
            return Path(p).as_uri()
        except:
            return p

    def __init__(
        self,
        bounding_box: Rectangle,
        url_to_mp3_file: str,
    ):
        super(SoundAnnotation, self).__init__(bounding_box=bounding_box)
        self._url_to_mp3_file: str = url_to_mp3_file

        # (Required) The type of annotation that this dictionary describes; shall be
        # Link for a link annotation.
        self[Name("Subtype")] = Name("Screen")

        # (Optional; PDF 1.1) An action that shall be performed when the link
        # annotation is activated (see 12.6, “Actions”).
        self[Name("A")] = Dictionary()
        self["A"][Name("Type")] = Name("Action")
        self["A"][Name("S")] = Name("Rendition")
        self["A"][Name("OP")] = bDecimal(0)

        # A/R
        self["A"][Name("R")] = Dictionary()
        self[Name("A")][Name("R")][Name("Type")] = Name("Rendition")
        self[Name("A")][Name("R")][Name("S")] = Name("MR")

        # A/R/C
        self[Name("A")][Name("R")][Name("C")] = Dictionary()
        self[Name("A")][Name("R")][Name("C")][Name("Type")] = Name("MediaClip")
        self[Name("A")][Name("R")][Name("C")][Name("S")] = Name("MCD")
        self[Name("A")][Name("R")][Name("C")][Name("CT")] = String("video/mp4")

        # A/R/C/D
        # fmt: off
        self[Name("A")][Name("R")][Name("C")][Name("D")] = Dictionary()
        self[Name("A")][Name("R")][Name("C")][Name("D")][Name("Type")] = Name("Filespec")
        self[Name("A")][Name("R")][Name("C")][Name("D")][Name("FS")] = Name("URL")
        self[Name("A")][Name("R")][Name("C")][Name("D")][Name("F")] = String(SoundAnnotation._make_canonical_file_path(self._url_to_mp3_file))
        # fmt: on

        # A/R/C/P
        # fmt: off
        self[Name("A")][Name("R")][Name("C")][Name("P")] = Dictionary()
        self[Name("A")][Name("R")][Name("C")][Name("P")][Name("TF")] = String("TEMPACCESS")
        # fmt: on

        # A/R
        self[Name("A")][Name("R")][Name("P")] = Dictionary()
        self[Name("A")][Name("R")][Name("P")][Name("BE")] = Dictionary()
        self[Name("A")][Name("R")][Name("P")][Name("BE")][Name("C")] = Boolean(True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A Bézier curve is a parametric curve used in computer graphics and related fields.
The curve, which is related to the Bernstein polynomial, is named after Pierre Bézier,
who used it in the 1960s for designing curves for the bodywork of Renault cars.
Other uses include the design of computer fonts and animation.
Bézier curves can be combined to form a Bézier spline, or generalized to higher dimensions to form Bézier surfaces.
The Bézier triangle is a special case of the latter.
"""

import typing
from decimal import Decimal

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.geometry.line_segment import LineSegment
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


def _bezier(p0, p1, p2, p3) -> typing.List[LineSegment]:
    pts = []
    ONE = Decimal(1)
    for t in [
        Decimal(x) for x in [0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
    ]:
        x = (
            (ONE - t) ** 3 * p0[0]
            + 3 * t * (ONE - t) ** 2 * p1[0]
            + 3 * t**2 * (ONE - t) * p2[0]
            + t**3 * p3[0]
        )
        y = (
            (ONE - t) ** 3 * p0[1]
            + 3 * t * (ONE - t) ** 2 * p1[1]
            + 3 * t**2 * (ONE - t) * p2[1]
            + t**3 * p3[1]
        )
        pts.append((x, y))

    # build List of LineSegments
    out: typing.List[LineSegment] = []
    for i in range(1, len(pts)):
        out.append(LineSegment(pts[i - 1][0], pts[i - 1][1], pts[i][0], pts[i][1]))

    # return
    return out


class AppendCubicBezierCurve1(CanvasOperator):
    """
    Append a cubic Bézier curve to the current path. The curve
    shall extend from the current point to the point (x3 , y3 ), using
    (x1 , y1 ) and (x2 , y2 ) as the Bézier control points (see 8.5.2.2,
    "Cubic Bézier Curves"). The new current point shall be
    (x3 , y3 ).
    """

    def __init__(self):
        super().__init__("c", 6)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invokes the c operator
        """
        assert isinstance(
            operands[0], Decimal
        ), "operand 0 of c operator must be of type Decimal"
        assert isinstance(
            operands[1], Decimal
        ), "operand 1 of c operator must be of type Decimal"
        assert isinstance(
            operands[2], Decimal
        ), "operand 2 of c operator must be of type Decimal"
        assert isinstance(
            operands[3], Decimal
        ), "operand 3 of c operator must be of type Decimal"
        assert isinstance(
            operands[4], Decimal
        ), "operand 4 of c operator must be of type Decimal"
        assert isinstance(
            operands[5], Decimal
        ), "operand 5 of c operator must be of type Decimal"

        # get graphic state
        canvas = canvas_stream_processor.get_canvas()
        gs = canvas.graphics_state

        # path should not be empty
        assert len(gs.path) > 0

        # build control points
        p0 = (gs.path[-1].x1, gs.path[-1].y1)
        p1 = (operands[0], operands[1])
        p2 = (operands[2], operands[3])
        p3 = (operands[4], operands[5])

        # append all paths
        for l in _bezier(p0, p1, p2, p3):
            gs.path.append(l)


class AppendCubicBezierCurve2(CanvasOperator):
    """
    Append a cubic Bézier curve to the current path. The curve
    shall extend from the current point to the point (x3 , y3 ), using
    the current point and (x2 , y2 ) as the Bézier control points (see
    8.5.2.2, "Cubic Bézier Curves"). The new current point shall
    be (x3 , y3 ).
    """

    def __init__(self):
        super().__init__("v", 4)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invokes the v operator
        """
        assert isinstance(
            operands[0], Decimal
        ), "operand 0 of v operator must be of type Decimal"
        assert isinstance(
            operands[1], Decimal
        ), "operand 1 of v operator must be of type Decimal"
        assert isinstance(
            operands[2], Decimal
        ), "operand 2 of v operator must be of type Decimal"
        assert isinstance(
            operands[3], Decimal
        ), "operand 3 of v operator must be of type Decimal"

        # get graphic state
        canvas = canvas_stream_processor.get_canvas()
        gs = canvas.graphics_state

        # path should not be empty
        assert len(gs.path) > 0

        # build control points
        p0 = (gs.path[-1].x1, gs.path[-1].y1)
        p1 = p0
        p2 = (operands[0], operands[1])
        p3 = (operands[2], operands[3])

        # append all paths
        for l in _bezier(p0, p1, p2, p3):
            gs.path.append(l)


class AppendCubicBezierCurve3(CanvasOperator):
    """
    Append a cubic Bézier curve to the current path. The curve
    shall extend from the current point to the point (x3 , y3 ), using
    (x1 , y1 ) and (x3 , y3 ) as the Bézier control points (see 8.5.2.2,
    "Cubic Bézier Curves"). The new current point shall be (x3 , y3 ).
    """

    def __init__(self):
        super().__init__("y", 4)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invokes the y operator
        """
        assert isinstance(
            operands[0], Decimal
        ), "operand 0 of y operator must be of type Decimal"
        assert isinstance(
            operands[1], Decimal
        ), "operand 1 of y operator must be of type Decimal"
        assert isinstance(
            operands[2], Decimal
        ), "operand 2 of y operator must be of type Decimal"
        assert isinstance(
            operands[3], Decimal
        ), "operand 3 of y operator must be of type Decimal"

        # get graphic state
        canvas = canvas_stream_processor.get_canvas()
        gs = canvas.graphics_state

        # path should not be empty
        assert len(gs.path) > 0

        # build control points
        p0 = (gs.path[-1].x1, gs.path[-1].y1)
        p1 = (operands[0], operands[1])
        p2 = (operands[2], operands[3])
        p3 = p2

        # append all paths
        for l in _bezier(p0, p1, p2, p3):
            gs.path.append(l)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Syllabification or syllabication, also known as hyphenation, is the separation of a word into syllables, whether spoken, written or signed.
A hyphenation algorithm is a set of rules, especially one codified for implementation in a computer program,
that decides at which points a word can be broken over two lines with a hyphen.
For example, a hyphenation algorithm might decide that impeachment can be broken as impeach-ment or im-peachment but not impe-achment.
One of the reasons for the complexity of the rules of word-breaking is that different "dialects" of English tend to differ on hyphenation[citation needed]:
American English tends to work on sound,
but British English tends to look to the origins of the word and then to sound.
There are also a large number of exceptions, which further complicates matters.
"""
import json
import typing
from pathlib import Path

from borb.datastructure.str_trie import Trie


class Hyphenation:
    """
    Syllabification or syllabication, also known as hyphenation, is the separation of a word into syllables, whether spoken, written or signed.
    A hyphenation algorithm is a set of rules, especially one codified for implementation in a computer program,
    that decides at which points a word can be broken over two lines with a hyphen.
    For example, a hyphenation algorithm might decide that impeachment can be broken as impeach-ment or im-peachment but not impe-achment.
    One of the reasons for the complexity of the rules of word-breaking is that different "dialects" of English tend to differ on hyphenation[citation needed]:
    American English tends to work on sound,
    but British English tends to look to the origins of the word and then to sound.
    There are also a large number of exceptions, which further complicates matters.
    """

    DO_NOT_HYPHENATE_BEFORE: int = 2
    DO_NOT_HYPHENATE_AFTER: int = -2

    def __init__(self, iso_language_code: str):
        self._patterns: Trie = Trie()
        self._min_prefix_length: int = 128
        self._max_prefix_length: int = 0
        self._min_suffix_length: int = 128
        self._max_suffix_length: int = 0
        self._exceptions: typing.List[str] = []

        # load data
        resources_dir: Path = Path(__file__).parent / "resources"
        hyphenation_pattern_file: Path = resources_dir / ("%s.json" % iso_language_code)
        assert hyphenation_pattern_file.exists(), (
            "No hyphenation file for %s" % iso_language_code
        )
        with open(hyphenation_pattern_file, "r") as json_file_handle:
            data = json.loads(json_file_handle.read())

            # load patterns
            for p in data["patterns"]:
                assert isinstance(p, str)
                for i in range(0, len(p)):
                    if not p[i].isdigit():
                        continue
                    prefix: str = "".join([c for c in p[0:i] if not c.isdigit()])
                    suffix: str = "".join([c for c in p[i:] if not c.isdigit()])
                    # keep track of prefix length
                    self._max_prefix_length = max(self._max_prefix_length, len(prefix))
                    self._min_prefix_length = min(self._min_prefix_length, len(prefix))
                    # keep track of suffix length
                    self._max_suffix_length = max(self._max_suffix_length, len(suffix))
                    self._min_suffix_length = min(self._min_suffix_length, len(suffix))
                    # insert into trie
                    digit: int = int(p[i])
                    self._patterns[prefix + "0" + suffix] = digit

            # load exceptions
            if "exceptions" in data:
                for e in data["exceptions"]:
                    self._exceptions.append(e)

    def hyphenate(self, s: str, hyphenation_character: str = chr(173)) -> str:
        """
        This function hyphenates the input word, inserting the hyphenation_character wherever the word *can* be split
        in syllables. This function returns the word with hyphenation_character inserted.
        """
        # don't spend effort on stupid input
        # fmt: off
        assert (len(hyphenation_character) == 1), "The hyphenation_character must be a single character str."
        # fmt: on
        if len(s) <= 4:
            return s

        # check exceptions
        for e in self._exceptions:
            e_raw_str: str = "".join([c for c in e if c.isalpha()])
            if e_raw_str == s:
                return "".join([c if c.isalpha() else hyphenation_character for c in e])

        # normal run of the algorithm
        s2: str = "." + s + "."
        hyphenation_info: typing.List[int] = [0 for _ in range(0, len(s2))]
        for i in range(2, len(s2)):
            for j in range(self._min_prefix_length, self._max_prefix_length + 1):
                if i - j < 0:
                    continue
                prefix: str = s2[(i - j) : i]
                for k in range(self._min_suffix_length, self._max_suffix_length + 1):
                    if j == 0 and k == 0:
                        continue
                    if i + k >= len(s2):
                        continue
                    suffix: str = s2[i : (i + k)]
                    value: typing.Optional[int] = self._patterns[prefix + "0" + suffix]
                    if value:
                        hyphenation_info[i] = max(hyphenation_info[i], value)
        s3: str = ""
        for i in range(1, len(hyphenation_info) - 1):
            # obey DO_NOT_HYPHENATE_BEFORE
            if (i - 1) <= Hyphenation.DO_NOT_HYPHENATE_BEFORE:
                s3 += s2[i]
                continue
            # obey DO_NOT_HYPHENATE_AFTER
            if (i - 1) >= len(s) + Hyphenation.DO_NOT_HYPHENATE_AFTER:
                s3 += s2[i]
                continue
            # do not allow split on last 2, or first 2 characters
            if hyphenation_info[i] % 2 == 1:
                s3 += hyphenation_character + s2[i]
            else:
                s3 += s2[i]
        # return
        return s3



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This class adds performs OCR and adds recognized text in an optional content group on the PDF.
    This enables the user to have a searchable PDF, whilst being able to turn on/off OCR features.
"""
# FIX: circular imports (1/2)
from __future__ import annotations

import datetime
import typing
import zlib
from decimal import Decimal
from pathlib import Path

# FIX: circular imports (2/2)
from typing import TYPE_CHECKING

from borb.datastructure.disjoint_set import disjointset
from borb.io.read.types import Decimal as bDecimal
from borb.io.read.types import Dictionary, List, Name, String
from borb.pdf.canvas.event.event_listener import Event
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.text.chunk_of_text import ChunkOfText
from borb.pdf.document.document import Document
from borb.pdf.page.page import Page
from borb.toolkit.ocr.ocr_image_render_event_listener import (
    OCREvent,
    OCRImageRenderEventListener,
)

EndDocumentEvent = type(None)
if TYPE_CHECKING:
    from borb.io.read.reference.xref_transformer import EndDocumentEvent


class OCRAsOptionalContentGroup(OCRImageRenderEventListener):
    """
    This class adds performs OCR and adds recognized text in an optional content group on the PDF.
    This enables the user to have a searchable PDF, whilst being able to turn on/off OCR features.
    """

    def __init__(
        self, tesseract_data_dir: Path, minimal_confidence: Decimal = Decimal(0.75)
    ):
        super(OCRAsOptionalContentGroup, self).__init__(
            tesseract_data_dir, minimal_confidence
        )
        self._ocr_events: typing.List[OCREvent] = []

    def _overlaps_vertically(self, r0: Rectangle, r1: Rectangle) -> bool:
        """
        This function returns True iff two Rectangle objects overlap vertically, False otherwise.
        """
        return (
            int(r0.get_y()) <= int(r1.get_y()) <= int(r0.get_y() + r0.get_height())
        ) or (int(r1.get_y()) <= int(r0.get_y()) <= int(r1.get_y() + r1.get_height()))

    def _add_ocr_optional_content_group(self, document: Document) -> None:
        # add OCProperties to Document (if needed)
        if "OCProperties" not in document["XRef"]["Trailer"]["Root"]:

            # The optional OCProperties entry in the document catalog (see 7.7.2, "Document Catalog") shall contain, when
            # present, the optional content properties dictionary, which contains a list of all the optional content groups in the
            # document, as well as information about the default and alternate configurations for optional content.
            document["XRef"]["Trailer"]["Root"][Name("OCProperties")] = Dictionary()

            # (Required) An array of indirect references to all the optional content
            # groups in the document (see 8.11.2, "Optional Content Groups"), in any
            # order. Every optional content group shall be included in this array.
            document["XRef"]["Trailer"]["Root"]["OCProperties"][Name("OCGs")] = List()

            # (Required) The default viewing optional content configuration dictionary
            # (see 8.11.4.3, "Optional Content Configuration Dictionaries").
            document["XRef"]["Trailer"]["Root"]["OCProperties"][
                Name("D")
            ] = Dictionary()

        # create an Optional Content Group Dictionary
        ocg_dict: Dictionary = Dictionary()
        ocg_dict[Name("Type")] = Name("OCG")
        ocg_dict[Name("Name")] = String("OCR by borb")
        ocg_dict[Name("Intent")] = Name("View")
        document["XRef"]["Trailer"]["Root"]["OCProperties"][Name("OCGs")].append(
            ocg_dict
        )

        # add to /Resources Dictionary of the Page
        now = datetime.datetime.now()
        ocr_layer_internal_name: str = "ocr%d%d%d" % (now.year, now.month, now.day)
        number_of_pages: typing.Optional[
            Decimal
        ] = document.get_document_info().get_number_of_pages()
        assert number_of_pages is not None
        for page_nr in range(0, int(number_of_pages)):
            page: Page = document.get_page(page_nr)
            if "Resources" not in page:
                page[Name("Resources")] = Dictionary
            if "Properties" not in page["Resources"]:
                page["Resources"][Name("Properties")] = Dictionary()
            page["Resources"]["Properties"][Name(ocr_layer_internal_name)] = ocg_dict

            # do nothing if no events are processed for this Page
            # fmt: off
            ocr_events_per_page: typing.List[OCREvent] = [x for x in self._ocr_events if x.get_page() == page]
            if len(ocr_events_per_page) == 0:
                continue
            # fmt: on

            # re-align events
            # fmt: off
            ds: disjointset = disjointset()
            for e in ocr_events_per_page:
                ds.add(e)
            for e1 in ocr_events_per_page:
                for e2 in ocr_events_per_page:
                    if e1 == e2:
                        continue
                    if self._overlaps_vertically(e1.get_bounding_box(), e2.get_bounding_box()):
                        ds.union(e1, e2)
            for es in ds.sets():
                avg_y: Decimal = Decimal(sum([x.get_bounding_box().get_y() for x in es]) / len(es))
                for e in es:
                    e.get_bounding_box().y = avg_y
            # fmt: on

            # change Page content stream
            # fmt: off
            page["Contents"][Name("DecodedBytes")] += ("\n/OC /%s BDC\n" % ocr_layer_internal_name).encode("latin1")
            for e in ocr_events_per_page:
                ChunkOfText(e.get_text(),
                            e.get_font(),
                            e.get_font_size(),
                            e.get_font_color()).paint(page, e.get_bounding_box())
            page["Contents"][Name("DecodedBytes")] += "\nEMC".encode("latin1")
            page["Contents"][Name("Bytes")] = zlib.compress(page["Contents"]["DecodedBytes"], 9)
            page["Contents"][Name("Length")] = bDecimal(len(page["Contents"][Name("Bytes")]))
            # fmt: on

    def _event_occurred(self, event: Event) -> None:
        super(OCRAsOptionalContentGroup, self)._event_occurred(event)
        # TODO: find a better way to solve this
        if event.__class__.__name__ == "EndDocumentEvent":
            self._end_document()

    def _end_document(self):
        if len(self._ocr_events) == 0:
            return
        document: Document = self._ocr_events[0].get_page().get_document()
        self._add_ocr_optional_content_group(document)

    def _ocr_text_occurred(self, event: OCREvent):
        self._ocr_events.append(event)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of WriteBaseTransformer is responsible for writing ASCII art in every PDF
"""
from pathlib import Path
from typing import Optional

from borb.io.read.types import AnyPDFType, Stream
from borb.io.write.transformer import Transformer, WriteTransformerState


class ASCIIArtTransformer(Transformer):
    """
    This implementation of WriteBaseTransformer is responsible for writing ASCII art in every PDF
    """

    def __init__(self):
        super().__init__()
        self._has_been_used: bool = False

    def can_be_transformed(self, any: AnyPDFType):
        """
        This function returns True once per Document (on the first Stream object) and embeds some ASCII art
        This is used to embed the current version in each Document
        """
        return isinstance(any, Stream) and not self._has_been_used

    def transform(
        self,
        object_to_transform: AnyPDFType,
        context: Optional[WriteTransformerState] = None,
    ):
        """
        This method writes ASCII art to a byte stream
        """
        # fmt: off
        assert (context is not None), "context must be defined to write ASCII art (borb meta-info)"
        assert (context.destination is not None), "context.destination must be defined to write ASCII art (borb meta-info)"
        assert isinstance(object_to_transform, Stream)
        # fmt: on

        f = Path(__file__).parent / "ascii_logo.txt"
        with open(f, "r") as logo_file_handle:
            ascii_logo = logo_file_handle.readlines()

        # append newline (if needed)
        if ascii_logo[-1][-1] != "\n":
            ascii_logo[-1] += "\n"

        # convert to latin1
        ascii_logo_bytes = [bytes("%    " + x, "utf8") for x in ascii_logo]

        self._has_been_used = True
        for x in ascii_logo_bytes:
            context.destination.write(x)
        context.destination.write(bytes("\n", "utf8"))

        self.get_root_transformer().transform(object_to_transform, context)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Square and circle annotations (PDF 1.3) shall display, respectively, a rectangle or an ellipse on the page. When
opened, they shall display a pop-up window containing the text of the associated note. The rectangle or ellipse
shall be inscribed within the annotation rectangle defined by the annotation dictionary’s Rect entry (see
Table 168).
"""
import typing
from decimal import Decimal

from borb.io.read.types import Name, List, Decimal as bDecimal
from borb.pdf.canvas.color.color import Color
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.annotation.annotation import Annotation


class CircleAnnotation(Annotation):
    """
    Square and circle annotations (PDF 1.3) shall display, respectively, a rectangle or an ellipse on the page. When
    opened, they shall display a pop-up window containing the text of the associated note. The rectangle or ellipse
    shall be inscribed within the annotation rectangle defined by the annotation dictionary’s Rect entry (see
    Table 168).
    """

    def __init__(
        self,
        bounding_box: Rectangle,
        fill_color: Color,
        stroke_color: Color,
        rectangle_difference: typing.Optional[
            typing.Tuple[Decimal, Decimal, Decimal, Decimal]
        ] = None,
    ):
        super(CircleAnnotation, self).__init__(
            bounding_box=bounding_box, color=stroke_color
        )

        # (Required) The type of annotation that this dictionary describes; shall be
        # Square or Circle for a square or circle annotation, respectively.
        self[Name("Subtype")] = Name("Circle")

        # (Optional) A border style dictionary (see Table 166) specifying the line
        # width and dash pattern that shall be used in drawing the rectangle or
        # ellipse.
        # The annotation dictionary’s AP entry, if present, shall take precedence
        # over the Rect and BS entries; see Table 168 and 12.5.5, “Appearance
        # Streams.”
        # self[Name("BS")] = None

        # (Optional; PDF 1.4) An array of numbers that shall be in the range 0.0 to
        # 1.0 and shall specify the interior color with which to fill the annotation’s
        # rectangle or ellipse. The number of array elements determines the colour
        # space in which the colour shall be defined
        if fill_color is not None:
            self[Name("IC")] = List().set_is_inline(True)  # type: ignore [attr-defined]
            self["IC"].append(bDecimal(fill_color.to_rgb().red))
            self["IC"].append(bDecimal(fill_color.to_rgb().green))
            self["IC"].append(bDecimal(fill_color.to_rgb().blue))

        # (Optional; PDF 1.5) A border effect dictionary describing an effect applied
        # to the border described by the BS entry (see Table 167).
        # self[Name("BE")] = None

        # (Optional; PDF 1.5) A set of four numbers that shall describe the
        # numerical differences between two rectangles: the Rect entry of the
        # annotation and the actual boundaries of the underlying square or circle.
        # Such a difference may occur in situations where a border effect
        # (described by BE) causes the size of the Rect to increase beyond that of
        # the square or circle.
        # The four numbers shall correspond to the differences in default user
        # space between the left, top, right, and bottom coordinates of Rect and
        # those of the square or circle, respectively. Each value shall be greater
        # than or equal to 0. The sum of the top and bottom differences shall be
        # less than the height of Rect, and the sum of the left and right differences
        # shall be less than the width of Rect.
        if rectangle_difference is not None:
            self[Name("RD")] = List().set_is_inline(True)  # type: ignore [attr-defined]
            self["RD"].append(bDecimal(rectangle_difference[0]))
            self["RD"].append(bDecimal(rectangle_difference[1]))
            self["RD"].append(bDecimal(rectangle_difference[2]))
            self["RD"].append(bDecimal(rectangle_difference[3]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Decompresses data encoded using a byte-oriented run-length
encoding algorithm, reproducing the original text or binary data
(typically monochrome image data, or any data that contains
frequent long runs of a single byte value).
"""
import logging

logger = logging.getLogger(__name__)


class RunLengthDecode:
    """
    Decompresses data encoded using a byte-oriented run-length
    encoding algorithm, reproducing the original text or binary data
    (typically monochrome image data, or any data that contains
    frequent long runs of a single byte value).
    """

    @staticmethod
    def decode(bytes_in: bytes) -> bytes:
        """
        Decompresses data encoded using a byte-oriented run-length
        encoding algorithm
        """

        # trivial case
        if len(bytes_in) == 0:
            return bytes_in

        # The RunLengthDecode filter decodes data that has been encoded in a simple byte-oriented format based on
        # run length. The encoded data shall be a sequence of runs, where each run shall consist of a length byte
        # followed by 1 to 128 bytes of data. If the length byte is in the range 0 to 127, the following length + 1 (1 to 128)
        # bytes shall be copied literally during decompression. If length is in the range 129 to 255, the following single
        # byte shall be copied 257 - length (2 to 128) times during decompression. A length value of 128 shall denote
        # EOF.
        bytes_out = bytearray()
        i: int = 0
        while i < len(bytes_in):
            b = bytes_in[i]
            # A length value of 128 shall denote EOF.
            if b == 128:
                break
            # If the length byte is in the range 0 to 127, the following length + 1 (1 to 128)
            # bytes shall be copied literally during decompression.
            length: int = 0
            if 0 <= b <= 127:
                length = b + 1
                i += 1
                for j in range(0, length):
                    bytes_out.append(bytes_in[i + j])
                i += length
                continue
            # If length is in the range 129 to 255, the following single
            # byte shall be copied 257 - length (2 to 128) times during decompression
            if 129 <= b <= 255:
                length = 257 - b
                i += 1
                for _ in range(0, length):
                    bytes_out.append(bytes_in[i])
                i += 1

        # return
        return bytes(bytes_out)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A pop-up annotation (PDF 1.3) displays text in a pop-up window for entry and editing. It shall not appear alone
but is associated with a markup annotation, its parent annotation, and shall be used for editing the parent’s text.
It shall have no appearance stream or associated actions of its own and shall be identified by the Popup entry
in the parent’s annotation dictionary (see Table 174). Table 183 shows the annotation dictionary entries specific
to this type of annotation.
"""
from borb.pdf.canvas.layout.annotation.annotation import Annotation


class PopupAnnotation(Annotation):
    """
    A pop-up annotation (PDF 1.3) displays text in a pop-up window for entry and editing. It shall not appear alone
    but is associated with a markup annotation, its parent annotation, and shall be used for editing the parent’s text.
    It shall have no appearance stream or associated actions of its own and shall be identified by the Popup entry
    in the parent’s annotation dictionary (see Table 174). Table 183 shows the annotation dictionary entries specific
    to this type of annotation.
    """

    def __init__(self):
        super(PopupAnnotation, self).__init__()
        # TODO



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of LayoutElement represents one uninterrupted block of text
"""
import typing
from decimal import Decimal

from borb.io.read.types import Dictionary, Name
from borb.pdf.canvas.color.color import Color, HexColor
from borb.pdf.canvas.font.font import Font
from borb.pdf.canvas.font.glyph_line import GlyphLine
from borb.pdf.canvas.font.simple_font.font_type_1 import StandardType1Font
from borb.pdf.canvas.font.simple_font.true_type_font import TrueTypeFont
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.layout_element import Alignment, LayoutElement
from borb.pdf.page.page import Page


class ChunkOfText(LayoutElement):
    """
    This implementation of LayoutElement represents one uninterrupted block of text
    """

    def __init__(
        self,
        text: str,
        font: typing.Union[Font, str] = "Helvetica",
        font_size: Decimal = Decimal(12),
        font_color: Color = HexColor("000000"),
        border_top: bool = False,
        border_right: bool = False,
        border_bottom: bool = False,
        border_left: bool = False,
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_bottom_left: Decimal = Decimal(0),
        border_color: Color = HexColor("000000"),
        border_width: Decimal = Decimal(1),
        padding_top: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        margin_top: typing.Optional[Decimal] = None,
        margin_right: typing.Optional[Decimal] = None,
        margin_bottom: typing.Optional[Decimal] = None,
        margin_left: typing.Optional[Decimal] = None,
        vertical_alignment: Alignment = Alignment.TOP,
        horizontal_alignment: Alignment = Alignment.LEFT,
        fixed_leading: typing.Optional[Decimal] = None,
        multiplied_leading: typing.Optional[Decimal] = None,
        background_color: typing.Optional[Color] = None,
    ):
        super().__init__(
            font_size=font_size,
            border_top=border_top,
            border_right=border_right,
            border_bottom=border_bottom,
            border_left=border_left,
            border_radius_top_left=border_radius_top_left,
            border_radius_top_right=border_radius_top_right,
            border_radius_bottom_right=border_radius_bottom_right,
            border_radius_bottom_left=border_radius_bottom_left,
            border_color=border_color,
            border_width=border_width,
            padding_top=padding_top,
            padding_right=padding_right,
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            margin_top=margin_top or Decimal(0),
            margin_right=margin_right or Decimal(0),
            margin_bottom=margin_bottom or Decimal(0),
            margin_left=margin_left or Decimal(0),
            vertical_alignment=vertical_alignment,
            horizontal_alignment=horizontal_alignment,
            background_color=background_color,
        )
        self._text = text

        # font information
        if isinstance(font, str):
            self._font: Font = StandardType1Font(font)
            assert self._font
        else:
            self._font = font
        self._font_color = font_color

        # leading
        if fixed_leading is None and multiplied_leading is None:
            multiplied_leading = Decimal(1.2)
        assert fixed_leading is not None or multiplied_leading is not None
        assert fixed_leading is None or fixed_leading > 0
        assert multiplied_leading is None or multiplied_leading > 0
        self._multiplied_leading: typing.Optional[Decimal] = multiplied_leading
        self._fixed_leading: typing.Optional[Decimal] = fixed_leading

    def get_text(self) -> str:
        """
        This function returns the text of this LayoutElement
        """
        return self._text

    def get_font(self) -> Font:
        """
        This function returns the Font of this LayoutElement
        """
        return self._font

    def get_font_color(self) -> Color:
        """
        This function returns the font Color of this LayoutElement
        """
        return self._font_color

    def _get_font_resource_name(self, font: Font, page: Page):
        # create resources if needed
        if "Resources" not in page:
            page[Name("Resources")] = Dictionary().set_parent(page)  # type: ignore [attr-defined]
        if "Font" not in page["Resources"]:
            page["Resources"][Name("Font")] = Dictionary()

        # insert font into resources
        font_resource_name = [
            k for k, v in page["Resources"]["Font"].items() if v == font
        ]
        if len(font_resource_name) > 0:
            return font_resource_name[0]
        else:
            font_index = len(page["Resources"]["Font"]) + 1
            page["Resources"]["Font"][Name("F%d" % font_index)] = font
            return Name("F%d" % font_index)

    def _write_text_bytes(self) -> str:
        hex_mode: bool = False
        # check glyphs
        for c in self._text:
            if ord(c) != self._font.unicode_to_character_identifier(c):
                hex_mode = True
                break
        # delegate
        if hex_mode or isinstance(self._font, TrueTypeFont):
            return self._write_text_bytes_in_hex()
        else:
            return self._write_text_bytes_in_ascii()

    def _write_text_bytes_in_hex(self) -> str:
        # decide whether we should write 1 byte, or 2 bytes
        font: Font = self._font
        use_four_bytes: bool = False
        if "Encoding" in font and font["Encoding"] in ["Identity-H", "Identity-V"]:
            use_four_bytes = True

        # write each cid
        sOut: str = ""
        for c in self._text:
            cid: typing.Optional[int] = self._font.unicode_to_character_identifier(c)
            assert cid is not None, "Font %s can not represent '%s'" % (
                self._font.get_font_name(),
                c,
            )
            hex_rep: str = hex(int(cid))[2:]
            hex_rep = self._pad_string_with_zeroes(hex_rep, 4 if use_four_bytes else 2)
            sOut += "".join(["<", hex_rep, ">"])
        return "".join(["[", sOut, "] TJ"])

    def _pad_string_with_zeroes(self, s: str, n: int = 2) -> str:
        while len(s) < n:
            s = "0" + s
        return s

    def _write_text_bytes_in_ascii(self) -> str:
        """
        This function escapes certain reserved characters in PDF strings.
        """
        sOut: str = ""
        for c in self._text:
            if c == "\r":
                sOut += "\\r"
            elif c == "\n":
                sOut += "\\n"
            elif c == "\t":
                sOut += "\\t"
            elif c == "\b":
                sOut += "\\b"
            elif c == "\f":
                sOut += "\\f"
            elif c in ["(", ")", "\\"]:
                sOut += "\\" + c
            elif 0 <= ord(c) < 8:
                sOut += "\\00" + oct(ord(c))[2:]
            elif 8 <= ord(c) < 32:
                sOut += "\\0" + oct(ord(c))[2:]
            else:
                sOut += c
        # default
        return "".join(["(", sOut, ") Tj"])

    #
    # RENDERING METHODS
    #

    def _get_content_box(self, available_space: Rectangle) -> Rectangle:
        # line height
        assert self._font_size is not None
        line_height: Decimal = self._font_size
        if self._multiplied_leading is not None:
            line_height *= self._multiplied_leading
        if self._fixed_leading is not None:
            line_height += self._fixed_leading

        # width
        w: Decimal = GlyphLine.from_str(
            self._text, self._font, self._font_size
        ).get_width_in_text_space()

        # return
        return Rectangle(
            available_space.get_x(),
            available_space.get_y() + available_space.get_height() - line_height,
            w,
            line_height,
        )

    def _paint_content_box(self, page: "Page", content_box: Rectangle) -> None:
        assert self._font
        rgb_color = self._font_color.to_rgb()

        # line height
        assert self._font_size is not None
        line_height: Decimal = self._font_size
        if self._multiplied_leading is not None:
            line_height *= self._multiplied_leading
        if self._fixed_leading is not None:
            line_height += self._fixed_leading

        # descent
        # TODO
        descent: Decimal = (self._font.get_descent() / Decimal(1000)) * self._font_size

        # determine what to write to page
        content = """
            q
            BT
            %f %f %f rg
            /%s %f Tf            
            %f 0 0 %f %f %f Tm            
            %s
            ET            
            Q
        """ % (
            float(rgb_color.red),  # rg
            float(rgb_color.green),  # rg
            float(rgb_color.blue),  # rg
            self._get_font_resource_name(self._font, page),  # Tf
            float(1),  # Tf
            float(self._font_size),  # Tm
            float(self._font_size),  # Tm
            float(content_box.get_x()),  # Tm
            float(
                content_box.get_y()
                + content_box.get_height()
                - self._font_size
                - descent
            ),  # Tm
            self._write_text_bytes(),  # Tj
        )
        page._append_to_content_stream(content)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This class provides static convenience methods for generating points of common line-art in electronic documents,
    such as arrows, rectangles, triangles, regular n-gons, stars, etc
"""
import math
import typing
from decimal import Decimal
from typing import Tuple

from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.line_art.blob_factory import BlobFactory


class LineArtFactory:
    """
    This class provides static convenience methods for generating points of common line-art in electronic documents,
    such as arrows, rectangles, triangles, regular n-gons, stars, etc
    """

    @staticmethod
    def EURion(
        bounding_box: Rectangle,
    ) -> typing.List[
        typing.Tuple[typing.Tuple[Decimal, Decimal], typing.Tuple[Decimal, Decimal]]
    ]:
        """
        The EURion constellation (also known as Omron rings or doughnuts) is a pattern of symbols incorporated into a number of secure documents
        such as banknotes and ownership title certificates designs worldwide since about 1996.
        It is added to help imaging software detect the presence of such a document in a digital image.
        Such software can then block the user from reproducing banknotes to prevent counterfeiting using colour photocopiers.
        According to research from 2004, the EURion constellation is used for colour photocopiers but probably not used in computer software.
        It has been reported that Adobe Photoshop will not allow editing of an image of a banknote,
        but in some versions this is believed to be due to a different, unknown digital watermark rather than the EURion constellation.
        :return:
        """
        # 269,  73 r 25 s 17
        #  85, 170 r 25 s 17
        # 237, 228 r 25 s 17
        # 475, 280 r 25 s 17
        # 263, 487 r 25 s 17
        line_segments = []
        for x, y in [(269, 73), (85, 170), (237, 228), (475, 280), (263, 487)]:

            # calculate points of a circle
            circle_segments = []
            for i in range(0, 360):
                px: Decimal = Decimal(math.cos(math.radians(i)) * 25 - (25 / 2) + x)
                py: Decimal = Decimal(math.sin(math.radians(i)) * 25 - (25 / 2) + y)
                circle_segments.append((px, py))

            # add segments
            line_segments.extend(
                [
                    (
                        circle_segments[i],
                        circle_segments[(i + 1) % len(circle_segments)],
                    )
                    for i in range(0, len(circle_segments))
                ]
            )

        # scale
        min_x: Decimal = min([min(l[0][0], l[1][0]) for l in line_segments])
        max_x: Decimal = max([max(l[0][0], l[1][0]) for l in line_segments])
        w: Decimal = max_x - min_x
        min_y: Decimal = min([min(l[0][1], l[1][1]) for l in line_segments])
        max_y: Decimal = max([max(l[0][1], l[1][1]) for l in line_segments])
        h: Decimal = max_y - min_y
        w_scale = bounding_box.get_width() / w
        h_scale = bounding_box.get_height() / h
        line_segments = [
            (
                (l[0][0] * w_scale, l[0][1] * h_scale),
                (l[1][0] * w_scale, l[1][1] * h_scale),
            )
            for l in line_segments
        ]

        # translate
        # TODO

        # return
        return line_segments

    @staticmethod
    def lissajours(
        bounding_box: Rectangle, x_frequency: int, y_frequency: int
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        A Lissajous curve /ˈlɪsəʒuː/, also known as Lissajous figure or Bowditch curve /ˈbaʊdɪtʃ/, is the graph of a system of parametric equations
        which describe complex harmonic motion.
        This family of curves was investigated by Nathaniel Bowditch in 1815, and later in more detail in 1857 by Jules Antoine Lissajous (for whom it has been named).
        The appearance of the figure is highly sensitive to the ratio x_frequency / y_frequency.
        For a ratio of 1, the figure is an ellipse, with special cases including circles.
        The visual form of these curves is often suggestive of a three-dimensional knot,
        and indeed many kinds of knots, including those known as Lissajous knots, project to the plane as Lissajous figures.
        """
        pts = []
        r = min(bounding_box.width, bounding_box.height) / Decimal(2)
        for i in range(0, 360 * x_frequency * y_frequency):
            x = Decimal(math.sin(math.radians(i * x_frequency))) * r
            y = Decimal(math.cos(math.radians(i * y_frequency))) * r
            pts.append((x, y))
        return pts

    @staticmethod
    def flowchart_process(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Represents a set of operations that changes value, form, or location of data. Represented as a rectangle.
        """
        return LineArtFactory.rectangle(bounding_box)

    @staticmethod
    def flowchart_decision(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Shows a conditional operation that determines which one of the two paths the program will take.
        The operation is commonly a yes/no question or true/false test. Represented as a diamond (rhombus).
        """
        return LineArtFactory.diamond(bounding_box)

    @staticmethod
    def flowchart_document(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Shows a printed document or report.
        """
        pts = []
        # build squiggly line
        arc_height = bounding_box.height / Decimal(8)
        half_arc_height = arc_height / Decimal(2)
        from_angle = 150
        to_angle = 390
        for i in range(from_angle, to_angle):
            x = Decimal(i) / Decimal(to_angle - from_angle) * bounding_box.width
            y = (
                Decimal(math.cos(math.radians(i))) * arc_height
                + half_arc_height
                + bounding_box.y
            )
            pts.append((x, y))
        # add rectangle top
        pa = pts[0]
        pb = pts[-1]
        pts = pts + [
            (pb[0], bounding_box.y + bounding_box.height),
            (pa[0], bounding_box.y + bounding_box.height),
            pa,
        ]
        # return
        return pts

    @staticmethod
    def flowchart_predefined_document(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Shows a predefined printed document or report.
        """
        pts = []
        # build squiggly line
        arc_height = bounding_box.height / Decimal(8)
        half_arc_height = arc_height / Decimal(2)
        from_angle = 150
        to_angle = 390
        for i in range(from_angle, to_angle):
            x = Decimal(i) / Decimal(to_angle - from_angle) * bounding_box.width
            y = (
                Decimal(math.cos(math.radians(i))) * arc_height
                + half_arc_height
                + bounding_box.y
            )
            pts.append((x, y))
            if i == int(from_angle + (to_angle - from_angle) / 10):
                pts.append((x, bounding_box.y + bounding_box.height))
                pts.append((x, y))
        # add rectangle top
        pa = pts[0]
        pb = pts[-1]
        pts = pts + [
            (pb[0], bounding_box.y + bounding_box.height),
            (pa[0], bounding_box.y + bounding_box.height),
            pa,
        ]
        # return
        return pts

    @staticmethod
    def flowchart_multiple_documents(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Represents multiple documents in a process
        """
        return []

    @staticmethod
    def flowchart_data(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Indicates the process of inputting and outputting data,
        as in entering data or displaying results. Represented as a rhomboid.
        """
        w25 = bounding_box.width * Decimal(0.25)
        return [
            (bounding_box.x, bounding_box.y),
            (bounding_box.x + w25, bounding_box.y + bounding_box.height),
            (bounding_box.x + bounding_box.width, bounding_box.y + bounding_box.height),
            (bounding_box.x + bounding_box.width - w25, bounding_box.y),
            (bounding_box.x, bounding_box.y),
        ]

    @staticmethod
    def flowchart_predefined_process(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Shows named process which is defined elsewhere.
        Represented as a rectangle with double-struck vertical edges.[14]
        """
        w10 = bounding_box.width * Decimal(0.1)
        return [
            (bounding_box.x, bounding_box.y),
            (bounding_box.x, bounding_box.y + bounding_box.height),
            (bounding_box.x + bounding_box.width, bounding_box.y + bounding_box.height),
            (bounding_box.x + bounding_box.width, bounding_box.y),
            # stripe (1)
            (bounding_box.x + bounding_box.width - w10, bounding_box.y),
            (
                bounding_box.x + bounding_box.width - w10,
                bounding_box.y + bounding_box.height,
            ),
            (bounding_box.x + bounding_box.width - w10, bounding_box.y),
            # stripe (2)
            (bounding_box.x + w10, bounding_box.y),
            (bounding_box.x + w10, bounding_box.y + bounding_box.height),
            (bounding_box.x + w10, bounding_box.y),
            (bounding_box.x, bounding_box.y),
        ]

    @staticmethod
    def flowchart_stored_data(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        This is a general data storage object used in the process flow as opposed to data which could be also stored on a hard drive,
        magnetic tape, memory card, of any other storage device.
        """
        pts_a = []
        pts_b = []
        for i in range(0, 180):
            # first curve
            x0 = Decimal(math.sin(math.radians(i + 180))) * (
                bounding_box.get_width() / Decimal(10)
            ) + (bounding_box.get_width() / Decimal(10))
            y0 = (
                Decimal(math.cos(math.radians(i + 180)))
                * (bounding_box.get_height() / Decimal(2))
                + bounding_box.get_y()
                + (bounding_box.get_height() / Decimal(2))
            )
            pts_a.append((x0, y0))
            # second curve
            x0 += bounding_box.get_width() * Decimal(0.9)
            pts_b.append((x0, y0))
        pts_b.reverse()
        return pts_a + pts_b + [pts_a[0]]

    @staticmethod
    def flowchart_internal_storage(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        This is a shape which is commonly found in programming flowcharts to illustrate the information stored in memory,
        as opposed to on a file. This shape is often referred to as the magnetic core memory of early computers;
        or the random access memory (RAM) as we call it today.

        """
        w = min(bounding_box.width, bounding_box.height)
        w10 = w * Decimal(0.1)
        return [
            (bounding_box.x, bounding_box.y),
            (bounding_box.x, bounding_box.y + w),
            (bounding_box.x + w, bounding_box.y + w),
            (bounding_box.x + w, bounding_box.y),
            (bounding_box.x, bounding_box.y),
            # cross
            (bounding_box.x + w10, bounding_box.y),
            (bounding_box.x + w10, bounding_box.y + w),
            (bounding_box.x, bounding_box.y + w),
            (bounding_box.x, bounding_box.y + w - w10),
            (bounding_box.x + w, bounding_box.y + w - w10),
            (bounding_box.x, bounding_box.y + w - w10),
            (bounding_box.x, bounding_box.y),
        ]

    @staticmethod
    def flowchart_sequential_data(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        This shape is supposed to look like a reel of tape with a small portion of tape extending from the reel.
        It represents magnetic tape storage which is also called sequential access storage.
        """
        pts = []
        r = min(bounding_box.width, bounding_box.height) / Decimal(2)
        mid_x = bounding_box.x + r
        mid_y = bounding_box.y + r
        for i in range(0, 320):
            x = Decimal(math.sin(math.radians(i + 180))) * r + mid_x
            y = Decimal(math.cos(math.radians(i + 180))) * r + mid_y
            pts.append((x, y))
        #
        y0 = pts[0][1]
        y350 = pts[-1][1]
        pts.append((r + mid_x, y350))
        pts.append((r + mid_x, y0))
        pts.append(pts[0])
        return pts

    @staticmethod
    def flowchart_direct_data(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Direct Data object in a process flow represents information stored which can be accessed directly.
        This object represents a computer's hard drive.
        """
        """
        This is a general data storage object used in the process flow as opposed to data which could be also stored on a hard drive,
        magnetic tape, memory card, of any other storage device.
        """
        return []

    @staticmethod
    def flowchart_manual_input(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Represented by quadrilateral, with the top irregularly sloping up from left to right,
        like the side view of a keyboard. Represents a step where a user is prompted to enter information manually.
        """
        h80 = bounding_box.height * Decimal(0.8)
        return [
            (bounding_box.x, bounding_box.y),
            (bounding_box.x, bounding_box.y + h80),
            (bounding_box.x + bounding_box.width, bounding_box.y + bounding_box.height),
            (bounding_box.x + bounding_box.width, bounding_box.y),
            (bounding_box.x, bounding_box.y),
        ]

    @staticmethod
    def flowchart_manual_operation(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Represented by a trapezoid with the longest parallel side at the top,
        to represent an operation or adjustment to process that can only be made manually.
        """
        h20 = bounding_box.height * Decimal(0.2)
        return [
            (bounding_box.x, bounding_box.y),
            (bounding_box.x, bounding_box.y + bounding_box.height),
            (
                bounding_box.x + bounding_box.width,
                bounding_box.y + bounding_box.height - h20,
            ),
            (bounding_box.x + bounding_box.width, bounding_box.y + h20),
            (bounding_box.x, bounding_box.y),
        ]

    @staticmethod
    def flowchart_card(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        This is the old IBM punched card. Each line of a program was punched into one IBM card.
        Then the cards were stacked in order and taken to a card reader.
        Usually the student would submit the cards and someone else would run them during the middle of the night,
        when the computer wasn't so busy. The output was printed on wide z-fold paper.
        If you made a mistake, you would have to resubmit the cards and wait another day.
        Large programs had stacks of cards several feet high. If you are using this shape, you need to update your hardware.
        """
        h825 = bounding_box.height * Decimal(0.825)
        w175 = bounding_box.width * Decimal(0.175)
        return [
            (bounding_box.x, bounding_box.y),
            (bounding_box.x, bounding_box.y + h825),
            (bounding_box.x + w175, bounding_box.y + bounding_box.height),
            (bounding_box.x + bounding_box.width, bounding_box.y + bounding_box.height),
            (bounding_box.x + bounding_box.width, bounding_box.y),
            (bounding_box.x, bounding_box.y),
        ]

    @staticmethod
    def flowchart_paper_tape(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        An outdated symbol rarely ever used in modern practices or process flows,
        but this shape could be used if you’re mapping out processes or input methods
        on much older computers and CNC machines.
        """
        pts_a = []
        pts_b = []
        # build squiggly line
        arc_height = bounding_box.height / Decimal(8)
        half_arc_height = arc_height / Decimal(2)
        from_angle = 150
        to_angle = 390
        for i in range(from_angle, to_angle):
            x = Decimal(i) / Decimal(to_angle - from_angle) * bounding_box.width
            ya = (
                Decimal(math.cos(math.radians(i))) * arc_height
                + half_arc_height
                + bounding_box.y
            )
            yb = (
                Decimal(math.cos(math.radians(i))) * arc_height
                + half_arc_height
                + bounding_box.y
                + bounding_box.height
            )
            pts_a.append((x, ya))
            pts_b.append((x, yb))
        # return
        return pts_a + [x for x in reversed(pts_b)] + [pts_a[0]]

    @staticmethod
    def flowchart_display(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Indicates a step that displays information.
        """
        pts_a = []
        r_major = bounding_box.height / Decimal(2)
        r_minor = bounding_box.width / Decimal(10)
        mid_x = bounding_box.x + bounding_box.width - r_minor
        mid_y = bounding_box.y + bounding_box.height / Decimal(2)
        for i in range(0, 180):
            # first curve
            x = Decimal(math.sin(math.radians(i))) * r_minor + mid_x
            y = Decimal(math.cos(math.radians(i))) * r_major + mid_y
            pts_a.append((x, y))
        return (
            pts_a
            + [
                (bounding_box.x + bounding_box.width / Decimal(10), bounding_box.y),
                (bounding_box.x, mid_y),
                (
                    bounding_box.x + bounding_box.width / Decimal(10),
                    bounding_box.y + bounding_box.height,
                ),
            ]
            + [pts_a[0]]
        )

    @staticmethod
    def flowchart_preparation(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Represented by an elongated hexagon,
        originally used for steps like setting a switch or initializing a routine.
        """
        half_height = bounding_box.height / Decimal(2)
        quarter_width = bounding_box.width / Decimal(4)
        return [
            (bounding_box.x, bounding_box.y + half_height),
            (bounding_box.x + quarter_width, bounding_box.y + bounding_box.height),
            (
                bounding_box.x + bounding_box.width - quarter_width,
                bounding_box.y + bounding_box.height,
            ),
            (bounding_box.x + bounding_box.width, bounding_box.y + half_height),
            (bounding_box.x + bounding_box.width - quarter_width, bounding_box.y),
            (bounding_box.x + quarter_width, bounding_box.y),
            (bounding_box.x, bounding_box.y + half_height),
        ]

    @staticmethod
    def flowchart_loop_limit(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Indicates the point at which a loop should stop.
        """
        w25 = bounding_box.width * Decimal(0.25)
        h75 = bounding_box.height * Decimal(0.75)
        return [
            (bounding_box.x, bounding_box.y),
            (bounding_box.x, bounding_box.y + h75),
            (bounding_box.x + w25, bounding_box.y + bounding_box.height),
            (
                bounding_box.x + bounding_box.width - w25,
                bounding_box.y + bounding_box.height,
            ),
            (bounding_box.x + bounding_box.width, bounding_box.y + h75),
            (bounding_box.x + bounding_box.width, bounding_box.y),
            (bounding_box.x, bounding_box.y),
        ]

    @staticmethod
    def flowchart_termination(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Indicates the beginning and ending of a program or sub-process.
        Represented as a stadium, oval or rounded (fillet) rectangle.
        They usually contain the word "Start" or "End", or another phrase signaling the start or end of a process,
        such as "submit inquiry" or "receive product".
        """
        pts_a = []
        pts_b = []
        r_major = bounding_box.height * Decimal(0.5)
        r_minor = bounding_box.width * Decimal(0.25)
        for i in range(0, 180):
            # first curve
            x = (
                Decimal(math.sin(math.radians(i))) * r_minor
                + bounding_box.width * Decimal(0.5)
                + bounding_box.x
            )
            y = Decimal(math.cos(math.radians(i))) * r_major + r_major + bounding_box.y
            pts_a.append((x, y))
            # second curve
            x = Decimal(math.sin(math.radians(i + 180))) * r_minor + bounding_box.x
            y = (
                Decimal(math.cos(math.radians(i + 180))) * r_major
                + r_major
                + bounding_box.y
            )
            pts_b.append((x, y))
        return pts_b + pts_a + [pts_b[0]]

    @staticmethod
    def flowchart_collate(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Indicates a step that orders information into a standard format.
        """
        return [
            (bounding_box.x, bounding_box.y),
            (bounding_box.x + bounding_box.width, bounding_box.y + bounding_box.height),
            (bounding_box.x, bounding_box.y + bounding_box.height),
            (bounding_box.x + bounding_box.width, bounding_box.y),
            (bounding_box.x, bounding_box.y),
        ]

    @staticmethod
    def flowchart_delay(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Represents a segment of delay in a process.
        It can be helpful to indicate the exact length of delay within the shape.
        """
        pts_a = []
        r_major = bounding_box.height * Decimal(0.5)
        r_minor = bounding_box.width * Decimal(0.25)
        for i in range(0, 180):
            x = (
                Decimal(math.sin(math.radians(i))) * r_minor
                + bounding_box.width * Decimal(0.5)
                + bounding_box.x
            )
            y = Decimal(math.cos(math.radians(i))) * r_major + r_major + bounding_box.y
            pts_a.append((x, y))
        return pts_a + [
            (bounding_box.x, pts_a[-1][1]),
            (bounding_box.x, bounding_box.y + bounding_box.height),
            pts_a[0],
        ]

    @staticmethod
    def flowchart_extract(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        The Extract shape involves removal of one or more specific sets of items from a set.
        For example, you could have a list of addresses and extract those that are within 10 miles of some location.
        """
        return [
            (bounding_box.x, bounding_box.y),
            (
                bounding_box.x + bounding_box.width / Decimal(2),
                bounding_box.y + bounding_box.height,
            ),
            (bounding_box.x + bounding_box.width, bounding_box.y),
            (bounding_box.x, bounding_box.y),
        ]

    @staticmethod
    def flowchart_merge(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        The Merge shape combines two or more sets of items into one set.
        """
        return [
            (bounding_box.x, bounding_box.y + bounding_box.height),
            (bounding_box.x + bounding_box.width / Decimal(2), bounding_box.y),
            (bounding_box.x + bounding_box.width, bounding_box.y + bounding_box.height),
            (bounding_box.x, bounding_box.y + bounding_box.height),
        ]

    @staticmethod
    def flowchart_or(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Just as described, this shape indicates that the process flow continues two paths or more.
        """
        pts = []
        r = min(bounding_box.width, bounding_box.height) / Decimal(2)
        mid_x = bounding_box.x + r
        mid_y = bounding_box.y + r
        for i in range(0, 360):
            x = Decimal(math.sin(math.radians(i))) * r + mid_x
            y = Decimal(math.cos(math.radians(i))) * r + mid_y
            pts.append((x, y))
            if i == 0 or i == 90:
                xb = Decimal(math.sin(math.radians(i + 180))) * r + mid_x
                yb = Decimal(math.cos(math.radians(i + 180))) * r + mid_y
                pts.append((xb, yb))
                pts.append((x, y))
        pts.append(pts[0])
        return pts

    @staticmethod
    def flowchart_sort(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Indicates a step that organizes a list of items into a sequence or sets based on some pre-determined criteria.
        """
        half_width = bounding_box.width / Decimal(2)
        half_height = bounding_box.height / Decimal(2)
        return [
            (bounding_box.x, bounding_box.y + half_height),
            (bounding_box.x + bounding_box.width, bounding_box.y + half_height),
            (bounding_box.x, bounding_box.y + half_height),
            (bounding_box.x + half_width, bounding_box.y + bounding_box.height),
            (bounding_box.x + bounding_box.width, bounding_box.y + half_height),
            (bounding_box.x + half_width, bounding_box.y),
            (bounding_box.x, bounding_box.y + half_height),
        ]

    @staticmethod
    def flowchart_summing_junction(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Indicates a point in the flowchart where multiple branches converge back into a single process.
        """
        pts = []
        r = min(bounding_box.width, bounding_box.height) / Decimal(2)
        mid_x = bounding_box.x + r
        mid_y = bounding_box.y + r
        for i in range(0, 360):
            x = Decimal(math.sin(math.radians(i))) * r + mid_x
            y = Decimal(math.cos(math.radians(i))) * r + mid_y
            pts.append((x, y))
            if i == 45 or i == 135:
                xb = Decimal(math.sin(math.radians(i + 180))) * r + mid_x
                yb = Decimal(math.cos(math.radians(i + 180))) * r + mid_y
                pts.append((xb, yb))
                pts.append((x, y))
        pts.append(pts[0])
        return pts

    @staticmethod
    def flowchart_database(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Indicates a list of information with a standard structure that allows for searching and sorting.
        """
        pts_a = []
        pts_b = []
        r_major = bounding_box.get_width() / Decimal(2)
        r_minor = r_major / Decimal(3)
        mid_x = bounding_box.x + r_major
        mid_y = bounding_box.y + bounding_box.get_height() - r_minor
        # first curve
        for i in range(90, 90 + 360 + 180):
            x = Decimal(math.sin(math.radians(i))) * r_major + mid_x
            y = Decimal(math.cos(math.radians(i))) * r_minor + mid_y
            pts_a.append((x, y))
        # second curve
        mid_y = bounding_box.y + r_minor
        for i in range(90, 270):
            x = Decimal(math.sin(math.radians(i))) * r_major + mid_x
            y = Decimal(math.cos(math.radians(i))) * r_minor + mid_y
            pts_b.append((x, y))
        pts_b.reverse()
        return pts_a + pts_b + [pts_a[0]]

    @staticmethod
    def flowchart_on_page_reference(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        This small circle (also known as Connector) indicates that the next (or previous) step is somewhere else on the drawing.
        This is particularly useful for large flowcharts where you would otherwise have to use a long connector,
        which can be hard to follow.
        """
        return LineArtFactory.circle(bounding_box)

    @staticmethod
    def flowchart_off_page_reference(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """ "
        A labeled connector for use when the target is on another page.
        Represented as a home plate-shaped pentagon.
        """
        return [
            (bounding_box.x, bounding_box.y),
            (bounding_box.x, bounding_box.y + bounding_box.height),
            (
                bounding_box.x + bounding_box.width * Decimal(0.5),
                bounding_box.y + bounding_box.height,
            ),
            (
                bounding_box.x + bounding_box.width,
                bounding_box.y + bounding_box.height * Decimal(0.5),
            ),
            (bounding_box.x + bounding_box.width * Decimal(0.5), bounding_box.y),
            (bounding_box.x, bounding_box.y),
        ]

    @staticmethod
    def flowchart_process_iso_9000(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        Represents a set of operations that changes value, form, or location of data. Represented as a rectangle.
        """
        w20 = bounding_box.width * Decimal(0.2)
        h50 = bounding_box.height * Decimal(0.5)
        return [
            (bounding_box.x, bounding_box.y),
            (bounding_box.x + w20, bounding_box.y + h50),
            (bounding_box.x, bounding_box.y + bounding_box.height),
            (
                bounding_box.x + bounding_box.width - w20,
                bounding_box.y + bounding_box.height,
            ),
            (bounding_box.x + bounding_box.width, bounding_box.y + h50),
            (bounding_box.x + bounding_box.width - w20, bounding_box.y),
            (bounding_box.x, bounding_box.y),
        ]

    @staticmethod
    def flowchart_transport(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        The Transport flowchart shape represents the step in the flowchart where information or materials are being transported from the process.
        """
        return []

    @staticmethod
    def smooth_dragon_curve(bounding_box: Rectangle, number_of_iterations: int = 10):
        """
        This function returns the coordinates for the Heighway dragon (also known as the Harter–Heighway dragon,
        or the Jurassic Park dragon) curve that fits in the given bounding box
        """
        points = LineArtFactory.dragon_curve(bounding_box, number_of_iterations)

        # smooth lines
        return BlobFactory.smooth_closed_polygon(points, 2)[:-6]

    @staticmethod
    def dragon_curve(
        bounding_box: Rectangle, number_of_iterations: int = 10
    ) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for the Heighway dragon (also known as the Harter–Heighway dragon,
        or the Jurassic Park dragon) curve that fits in the given bounding box
        """
        seq: typing.List[int] = [1]
        for _ in range(0, number_of_iterations):
            seq.append(1)
            m: int = int(len(seq) / 2) - 1
            for i, v in enumerate(seq[0:-1]):
                if i == m:
                    seq.append(1 - v)
                else:
                    seq.append(v)
        step_size = Decimal(10)
        direction: int = 0
        x: Decimal = Decimal(0)
        y: Decimal = Decimal(0)
        points: typing.List[Tuple[Decimal, Decimal]] = []
        for turn in seq:
            # go forward
            if direction == 0:
                y += step_size
            elif direction == 1:
                x += step_size
            elif direction == 2:
                y -= step_size
            elif direction == 3:
                x -= step_size
            # store point
            points.append((x, y))
            # make turn
            if turn == 0:
                direction = (direction + 1) % 4
            elif turn == 1:
                direction = (direction + 3) % 4

        # determine width/height
        w: Decimal = max([x[0] for x in points]) - min([x[0] for x in points])
        h: Decimal = max([x[1] for x in points]) - min([x[1] for x in points])

        # scale everything
        w_scale: Decimal = bounding_box.width / w
        h_scale: Decimal = bounding_box.height / h
        points = [(x[0] * w_scale, x[1] * h_scale) for x in points]

        # translate everything
        x_delta: Decimal = x - bounding_box.x
        y_delta: Decimal = y - bounding_box.y
        points = [(x[0] - x_delta, x[1] - y_delta) for x in points]

        return points

    @staticmethod
    def cross(bounding_box: Rectangle) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a cross that matches the given bounding box
        """
        return [
            (bounding_box.x, bounding_box.y + bounding_box.height * Decimal(0.66)),
            (
                bounding_box.x + bounding_box.width * Decimal(0.33),
                bounding_box.y + bounding_box.height * Decimal(0.66),
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.33),
                bounding_box.y + bounding_box.height,
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.66),
                bounding_box.y + bounding_box.height,
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.66),
                bounding_box.y + bounding_box.height * Decimal(0.66),
            ),
            (
                bounding_box.x + bounding_box.width,
                bounding_box.y + bounding_box.height * Decimal(0.66),
            ),
            (
                bounding_box.x + bounding_box.width,
                bounding_box.y + bounding_box.height * Decimal(0.33),
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.66),
                bounding_box.y + bounding_box.height * Decimal(0.33),
            ),
            (bounding_box.x + bounding_box.width * Decimal(0.66), bounding_box.y),
            (bounding_box.x + bounding_box.width * Decimal(0.33), bounding_box.y),
            (
                bounding_box.x + bounding_box.width * Decimal(0.33),
                bounding_box.y + bounding_box.height * Decimal(0.33),
            ),
            (bounding_box.x, bounding_box.y + bounding_box.height * Decimal(0.33)),
            # repeat first point to explicitly close shape
            (bounding_box.x, bounding_box.y + bounding_box.height * Decimal(0.66)),
        ]

    @staticmethod
    def cartoon_diamond(
        bounding_box: Rectangle,
    ) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a cartoon diamond that matches the given bounding box
        """
        top_ratio: Decimal = Decimal(0.75)
        return [
            (bounding_box.x + bounding_box.width * Decimal(0.5), bounding_box.y),
            (bounding_box.x, bounding_box.y + bounding_box.height * top_ratio),
            (
                bounding_box.x + bounding_box.width * Decimal(0.2),
                bounding_box.y + bounding_box.height,
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.4),
                bounding_box.y + bounding_box.height,
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.33),
                bounding_box.y + bounding_box.height * top_ratio,
            ),
            (bounding_box.x + bounding_box.width * Decimal(0.5), bounding_box.y),
            (
                bounding_box.x + bounding_box.width * Decimal(0.66),
                bounding_box.y + bounding_box.height * top_ratio,
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.6),
                bounding_box.y + bounding_box.height,
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.4),
                bounding_box.y + bounding_box.height,
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.8),
                bounding_box.y + bounding_box.height,
            ),
            (
                bounding_box.x + bounding_box.width,
                bounding_box.y + bounding_box.height * top_ratio,
            ),
            (bounding_box.x, bounding_box.y + bounding_box.height * top_ratio),
            (
                bounding_box.x + bounding_box.width,
                bounding_box.y + bounding_box.height * top_ratio,
            ),
            # repeat first point to explicitly close shape
            (bounding_box.x + bounding_box.width * Decimal(0.5), bounding_box.y),
        ]

    @staticmethod
    def rectangle(bounding_box: Rectangle) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a rectangle that matches the given bounding box
        """
        return [
            (bounding_box.x, bounding_box.y),
            (bounding_box.x + bounding_box.width, bounding_box.y),
            (bounding_box.x + bounding_box.width, bounding_box.y + bounding_box.height),
            (bounding_box.x, bounding_box.y + bounding_box.height),
            # repeat first point to explicitly close shape
            (bounding_box.x, bounding_box.y),
        ]

    @staticmethod
    def right_angled_triangle(
        bounding_box: Rectangle,
    ) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a right-angled triangle that fits in the given bounding box
        """
        return [
            (bounding_box.x, bounding_box.y),
            (bounding_box.x + bounding_box.width, bounding_box.y),
            (bounding_box.x, bounding_box.y + bounding_box.height),
            # repeat first point to explicitly close shape
            (bounding_box.x, bounding_box.y),
        ]

    @staticmethod
    def regular_n_gon(
        bounding_box: Rectangle, n: int
    ) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a regular n-gon that fits in the given bounding box
        """
        r = min(bounding_box.width, bounding_box.height) / Decimal(2)
        mid_x = bounding_box.x + r
        mid_y = bounding_box.y + r
        points = []
        for i in range(0, 360, int(360 / n)):
            x = Decimal(math.sin(math.radians(i))) * r + mid_x
            y = Decimal(math.cos(math.radians(i))) * r + mid_y
            points.append((x, y))
        points.append(points[0])
        return points

    @staticmethod
    def isosceles_triangle(
        bounding_box: Rectangle,
    ) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for an isosceles triangle that fits in the given bounding box
        """
        return LineArtFactory.regular_n_gon(bounding_box, 3)

    @staticmethod
    def parallelogram(bounding_box: Rectangle) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a parallelogram that fits in the given bounding box
        """
        return [
            (bounding_box.x, bounding_box.y),
            (
                bounding_box.x + bounding_box.width * Decimal(0.75),
                bounding_box.y,
            ),
            (
                bounding_box.x + bounding_box.width,
                bounding_box.y + bounding_box.height,
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.25),
                bounding_box.y + bounding_box.height,
            ),
            # repeat first point to explicitly close shape
            (bounding_box.x, bounding_box.y),
        ]

    @staticmethod
    def trapezoid(bounding_box: Rectangle) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a trapezoid that fits in the given bounding box
        """
        return [
            (bounding_box.x, bounding_box.y),
            (bounding_box.x + bounding_box.width, bounding_box.y),
            (
                bounding_box.x + bounding_box.width * Decimal(0.75),
                bounding_box.y + bounding_box.height,
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.25),
                bounding_box.y + bounding_box.height,
            ),
            # repeat first point to explicitly close shape
            (bounding_box.x, bounding_box.y),
        ]

    @staticmethod
    def diamond(bounding_box: Rectangle) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a diomond that fits in the given bounding box
        """
        HALF: Decimal = Decimal(0.5)
        return [
            (
                bounding_box.x + bounding_box.width * HALF,
                bounding_box.y,
            ),
            (
                bounding_box.x + bounding_box.width,
                bounding_box.y + bounding_box.height * HALF,
            ),
            (
                bounding_box.x + bounding_box.width * HALF,
                bounding_box.y + bounding_box.height,
            ),
            (
                bounding_box.x,
                bounding_box.y + bounding_box.height * HALF,
            ),
            # repeat first point to explicitly close shape
            (
                bounding_box.x + bounding_box.width * HALF,
                bounding_box.y,
            ),
        ]

    @staticmethod
    def pentagon(bounding_box: Rectangle) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a pentagon that fits in the given bounding box
        """
        return LineArtFactory.regular_n_gon(bounding_box, 5)

    @staticmethod
    def hexagon(bounding_box: Rectangle) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a hexagon that fits in the given bounding box
        """
        return LineArtFactory.regular_n_gon(bounding_box, 6)

    @staticmethod
    def heptagon(bounding_box: Rectangle) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a heptagon that fits in the given bounding box
        """
        return LineArtFactory.regular_n_gon(bounding_box, 7)

    @staticmethod
    def octagon(bounding_box: Rectangle) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for an octagon that fits in the given bounding box
        """
        return LineArtFactory.regular_n_gon(bounding_box, 8)

    @staticmethod
    def circle(
        bounding_box: Rectangle,
    ) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a circle that fits in the given bounding box
        """
        r = Decimal(min(bounding_box.width, bounding_box.height)) / Decimal(2)
        mid_x = bounding_box.x + r
        mid_y = bounding_box.y + r
        points: typing.List[Tuple[Decimal, Decimal]] = []
        for i in range(0, 360):
            x = Decimal(math.sin(math.radians(i))) * r + mid_x
            y = Decimal(math.cos(math.radians(i))) * r + mid_y
            points.append((x, y))
        return points

    @staticmethod
    def fraction_of_circle(
        bounding_box: Rectangle,
        fraction: Decimal,
    ) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a circle-slice that fits in the given bounding box
        """
        r = Decimal(min(bounding_box.width, bounding_box.height)) / Decimal(2)
        mid_x = bounding_box.x + r
        mid_y = bounding_box.y + r
        points: typing.List[Tuple[Decimal, Decimal]] = []
        for i in range(0, int(360 * float(fraction))):
            x = Decimal(math.sin(math.radians(i))) * r + mid_x
            y = Decimal(math.cos(math.radians(i))) * r + mid_y
            points.append((x, y))
        points.append((mid_x, mid_y))
        # repeat first point to explicitly close shape
        points.append(points[0])
        return points

    @staticmethod
    def three_quarters_of_circle(
        bounding_box: Rectangle,
    ) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a circle-slice of 270 degrees that fits in the given bounding box
        """
        return LineArtFactory.fraction_of_circle(bounding_box, Decimal(0.75))

    @staticmethod
    def half_of_circle(bounding_box: Rectangle) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a circle-slice of 180 degrees that fits in the given bounding box
        """
        return LineArtFactory.fraction_of_circle(bounding_box, Decimal(0.5))

    @staticmethod
    def droplet(bounding_box: Rectangle) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a droplet that fits in the given bounding box
        """
        r = Decimal(min(bounding_box.width, bounding_box.height)) / Decimal(2)
        mid_x = bounding_box.x + r
        mid_y = bounding_box.y + r
        points: typing.List[Tuple[Decimal, Decimal]] = []
        for i in range(0, 270):
            x = Decimal(math.sin(math.radians(i))) * r + mid_x
            y = Decimal(math.cos(math.radians(i))) * r + mid_y
            points.append((x, y))
        points.append((points[-1][0], points[0][1]))
        # repeat first point to explicitly close shape
        points.append(points[0])
        return points

    @staticmethod
    def four_pointed_star(
        bounding_box: Rectangle,
    ) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a four point star that fits in the given bounding box
        """
        return LineArtFactory.n_pointed_star(bounding_box, 4)

    @staticmethod
    def five_pointed_star(
        bounding_box: Rectangle,
    ) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a five point star that fits in the given bounding box
        """
        return LineArtFactory.n_pointed_star(bounding_box, 5)

    @staticmethod
    def six_pointed_star(
        bounding_box: Rectangle,
    ) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a six point star that fits in the given bounding box
        """
        return LineArtFactory.n_pointed_star(bounding_box, 6)

    @staticmethod
    def n_pointed_star(
        bounding_box: Rectangle, n: int
    ) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for an n-point star that fits in the given bounding box
        """
        assert n >= 3, "An n-pointed star must have at least 3 points"
        r = min(bounding_box.width, bounding_box.height) / Decimal(2)
        mid_x = bounding_box.x + r
        mid_y = bounding_box.y + r
        inner_radius = r * Decimal(0.39)
        points: typing.List[Tuple[Decimal, Decimal]] = []
        for i in range(0, 360, int(360 / n)):
            # outer point
            x = Decimal(math.sin(math.radians(i))) * r + mid_x
            y = Decimal(math.cos(math.radians(i))) * r + mid_y
            points.append((x, y))
            # inner point
            half_angle = int(360 / (2 * n))
            x = Decimal(math.sin(math.radians(i + half_angle))) * inner_radius + mid_x
            y = Decimal(math.cos(math.radians(i + half_angle))) * inner_radius + mid_y
            points.append((x, y))
        points.append(points[0])
        return points

    @staticmethod
    def arrow_left(bounding_box: Rectangle) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for an arrow pointing left that fits in the given bounding box
        """
        return [
            (
                bounding_box.x,
                bounding_box.y + bounding_box.width * Decimal(0.5),
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.39),
                bounding_box.y,
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.39),
                bounding_box.y + bounding_box.height * Decimal(0.2),
            ),
            (
                bounding_box.x + bounding_box.width,
                bounding_box.y + bounding_box.height * Decimal(0.2),
            ),
            (
                bounding_box.x + bounding_box.width,
                bounding_box.y + bounding_box.height * Decimal(0.8),
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.39),
                bounding_box.y + bounding_box.height * Decimal(0.8),
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.39),
                bounding_box.y + bounding_box.height,
            ),
            # repeat first point to explicitly close shape
            (
                bounding_box.x,
                bounding_box.y + bounding_box.width * Decimal(0.5),
            ),
        ]

    @staticmethod
    def arrow_right(bounding_box: Rectangle) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for an arrow pointing right that fits in the given bounding box
        """
        return [
            (
                bounding_box.x + bounding_box.width,
                bounding_box.y + bounding_box.width * Decimal(0.5),
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.61),
                bounding_box.y + bounding_box.height,
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.61),
                bounding_box.y + bounding_box.height * Decimal(0.8),
            ),
            (
                bounding_box.x,
                bounding_box.y + bounding_box.height * Decimal(0.8),
            ),
            (
                bounding_box.x,
                bounding_box.y + bounding_box.height * Decimal(0.2),
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.61),
                bounding_box.y + bounding_box.height * Decimal(0.2),
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.61),
                bounding_box.y,
            ),
            # repeat first point to explicitly close shape
            (
                bounding_box.x + bounding_box.width,
                bounding_box.y + bounding_box.width * Decimal(0.5),
            ),
        ]

    @staticmethod
    def arrow_up(bounding_box: Rectangle) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for an arrow pointing up that fits in the given bounding box
        """
        return [
            (
                bounding_box.x + bounding_box.width * Decimal(0.5),
                bounding_box.y + bounding_box.height,
            ),
            (bounding_box.x, bounding_box.y + bounding_box.height * Decimal(0.61)),
            (
                bounding_box.x + bounding_box.width * Decimal(0.2),
                bounding_box.y + bounding_box.height * Decimal(0.61),
            ),
            (bounding_box.x + bounding_box.width * Decimal(0.2), bounding_box.y),
            (bounding_box.x + bounding_box.width * Decimal(0.8), bounding_box.y),
            (
                bounding_box.x + bounding_box.width * Decimal(0.8),
                bounding_box.y + bounding_box.height * Decimal(0.61),
            ),
            (
                bounding_box.x + bounding_box.width,
                bounding_box.y + bounding_box.height * Decimal(0.61),
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.5),
                bounding_box.y + bounding_box.height,
            ),
        ]

    @staticmethod
    def arrow_down(bounding_box: Rectangle) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for an arrow pointing down that fits in the given bounding box
        """
        return [
            (bounding_box.x + bounding_box.width * Decimal(0.5), bounding_box.y),
            (
                bounding_box.x + bounding_box.width,
                bounding_box.y + bounding_box.height * Decimal(0.39),
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.8),
                bounding_box.y + bounding_box.height * Decimal(0.39),
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.8),
                bounding_box.y + bounding_box.height,
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.2),
                bounding_box.y + bounding_box.height,
            ),
            (
                bounding_box.x + bounding_box.width * Decimal(0.2),
                bounding_box.y + bounding_box.height * Decimal(0.39),
            ),
            (bounding_box.x, bounding_box.y + bounding_box.height * Decimal(0.39)),
            (bounding_box.x + bounding_box.width * Decimal(0.5), bounding_box.y),
        ]

    @staticmethod
    def heart(bounding_box: Rectangle) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a heart that fits in the given bounding box
        """
        r = min(bounding_box.width, bounding_box.height) / Decimal(2)
        points: typing.List[Tuple[Decimal, Decimal]] = []
        # first arc
        for i in range(0, 180):
            x = (
                Decimal(math.sin(math.radians(i - 90))) * r * Decimal(0.5)
                + bounding_box.x
                + r * Decimal(0.5)
            )
            y = (
                Decimal(math.cos(math.radians(i - 90))) * r * Decimal(0.5)
                + bounding_box.y
                + r
            )
            points.append((x, y))
        midpoint = points[-1]
        # second arc
        for i in range(0, 180):
            x = (
                Decimal(math.sin(math.radians(i - 90))) * r * Decimal(0.5)
                + bounding_box.x
                + r * Decimal(1.5)
            )
            y = (
                Decimal(math.cos(math.radians(i - 90))) * r * Decimal(0.5)
                + bounding_box.y
                + r
            )
            points.append((x, y))
        # triangle
        points.append((midpoint[0], bounding_box.y))
        points.append(points[0])
        return points

    @staticmethod
    def sticky_note(bounding_box: Rectangle) -> typing.List[Tuple[Decimal, Decimal]]:
        """
        This function returns the coordinates for a sticky note that fits in the given bounding box
        """
        turn_up: Decimal = Decimal(0.10)
        turn_up_inv: Decimal = Decimal(1) - turn_up
        return [
            (bounding_box.x, bounding_box.y),
            (bounding_box.x, bounding_box.y + bounding_box.height),
            (bounding_box.x + bounding_box.width, bounding_box.y + bounding_box.height),
            (
                bounding_box.x + bounding_box.width,
                bounding_box.y + bounding_box.height * turn_up,
            ),
            (
                bounding_box.x + bounding_box.width * turn_up_inv,
                bounding_box.y + bounding_box.height * turn_up,
            ),
            (bounding_box.x + bounding_box.width * turn_up_inv, bounding_box.y),
            (
                bounding_box.x + bounding_box.width,
                bounding_box.y + bounding_box.height * turn_up,
            ),
            (bounding_box.x + bounding_box.width * turn_up_inv, bounding_box.y),
            # repeat first point to explicitly close shape
            (bounding_box.x, bounding_box.y),
        ]
